"""
This module provides the Graph class, which represents a graph structure
extending the capabilities of BaseStructure to include graph-specific
properties and methods, such as detecting heads, acyclicity, and conversion to
NetworkX for visualization.
"""

from collections import deque
from typing import Any

from lionagi.core.generic import Node
from lionagi.core.generic import BaseStructure


class Graph(BaseStructure):
    """
    Represents a graph structure extending the capabilities of BaseStructure.

    Includes graph-specific properties and methods, such as detecting heads,
    acyclicity, and conversion to NetworkX for visualization.
    """

    def get_heads(self):
        return [
            node
            for node in self.internal_nodes.values()
            if not node.relations.pointed_by
        ]

    @property
    def acyclic(self) -> bool:
        """
        Checks if the graph is acyclic.

        An acyclic graph contains no cycles and can be represented as a directed
        acyclic graph (DAG).

        Returns:
            bool: True if the graph is acyclic, False otherwise.
        """
        node_ids = list(self.internal_nodes.keys())
        check_deque = deque(node_ids)
        check_dict = {key: 0 for key in node_ids}  # 0: not visited, 1: temp, 2: perm

        def visit(key):
            if check_dict[key] == 2:
                return True
            elif check_dict[key] == 1:
                return False

            check_dict[key] = 1

            points_to_edges = self.internal_nodes[key].relations.points_to

            for _, edge in points_to_edges.items():
                check = visit(edge.tail)
                if not check:
                    return False

            check_dict[key] = 2
            return True

        while check_deque:
            key = check_deque.pop()
            check = visit(key)
            if not check:
                return False
        return True

    def to_networkx(self, **kwargs) -> Any:
        """
        Converts the graph into a NetworkX graph object.

        The NetworkX graph object can be used for further analysis or
        visualization.

        Args:
            **kwargs: Additional keyword arguments to pass to the NetworkX graph
                constructor.

        Returns:
            Any: A NetworkX graph object representing the current graph
            structure.
        """
        from lionagi.libs import SysUtil

        SysUtil.check_import("networkx")

        from networkx import DiGraph

        g = DiGraph(**kwargs)
        for node_id, node in self.internal_nodes.items():
            node_info = node.to_dict()
            node_info.pop("id_")
            node_info.update({"class_name": node.class_name()})
            g.add_node(node_id, **node_info)

        for _edge in list(self.internal_edges.values()):
            edge_info = _edge.to_dict()
            edge_info.pop("id_")
            edge_info.update({"class_name": _edge.__class__.__name__})
            source_node_id = edge_info.pop("head")
            target_node_id = edge_info.pop("tail")
            g.add_edge(source_node_id, target_node_id, **edge_info)

        return g

    def display(self, **kwargs):
        """
        Displays the graph using NetworkX's drawing capabilities.

        This method requires NetworkX and a compatible plotting library (like
        matplotlib) to be installed.

        Args:
            **kwargs: Additional keyword arguments to pass to the NetworkX graph
                constructor.
        """
        from lionagi.libs import SysUtil

        SysUtil.check_import("networkx")

        from networkx import draw

        g = self.to_networkx(**kwargs)
        draw(g)



"""This module provides base components used in lionagi"""

from abc import ABC
from functools import singledispatchmethod
from typing import Any, TypeVar

from pydantic import AliasChoices, BaseModel, Field, ValidationError
from pandas import DataFrame, Series

from lionagi.libs import SysUtil, func_call, convert, ParseUtil, nested


T = TypeVar("T")


class BaseComponent(BaseModel, ABC):
    """
    Base class for creating component models.

    Attributes:
        id_ (str): A 32-char unique hash identifier for the node.
        timestamp (str): The timestamp of when the node was created.
    """

    id_: str = Field(
        title="ID",
        default_factory=SysUtil.create_id,
        validation_alias=AliasChoices("node_id", "ID", "id"),
        description="A 32-char unique hash identifier for the node.",
    )

    timestamp: str = Field(
        default_factory=lambda: SysUtil.get_timestamp(sep=None),
        description="The timestamp of when the node was created.",
    )

    class Config:
        """Model configuration settings."""

        extra = "allow"
        arbitrary_types_allowed = True
        populate_by_name = True
        validate_assignment = True
        validate_return = True
        str_strip_whitespace = True

    @classmethod
    def class_name(cls) -> str:
        """
        Retrieve the name of the class.

        Returns:
            str: The name of the class.
        """
        return cls.__name__

    @property
    def _field_annotations(self) -> dict:
        """
        Return the annotations for each field in the model.

        Returns:
            dict: A dictionary mapping field names to their annotations.
        """
        return self._get_field_annotation(list(self.model_fields.keys()))

    def _get_field_attr(self, k: str, attr: str, default=False) -> Any:
        """
        Get the value of a field attribute.

        Args:
            k (str): The field name.
            attr (str): The attribute name.
            default (Any): Default value to return if the attribute is not
                found.

        Returns:
            Any: The value of the field attribute, or the default value if not
                found.

        Raises:
            ValueError: If the field does not have the specified attribute.
        """
        try:
            if not self._field_has_attr(k, attr):
                raise ValueError(f"field {k} has no attribute {attr}")
            field = self.model_fields[k]
            a = getattr(field, attr)
            if not a:
                try:
                    a = field.json_schema_extra[attr]
                    return a
                except Exception:
                    return None
            return a
        except Exception as e:
            if default is not False:
                return default
            raise e

    @singledispatchmethod
    def _get_field_annotation(self, field_name: str) -> Any:
        """
        Get the annotation for a field.

        Args:
            field_name (str): The name of the field.

        Raises:
            TypeError: If the field_name is of an unsupported type.
        """
        raise TypeError(f"Unsupported type {type(field_name)}")

    @_get_field_annotation.register(str)
    def _(self, field_name) -> dict[str, Any]:
        """
        Get the annotation for a field as a dictionary.

        Args:
            field_name (str): The name of the field.

        Returns:
            dict[str, Any]: A dictionary mapping the field name to its
                annotation.
        """
        dict_ = {field_name: self.model_fields[field_name].annotation}
        for k, v in dict_.items():
            if "|" in str(v):
                v = str(v)
                v = v.split("|")
                dict_[k] = func_call.lcall(v, convert.strip_lower)
            else:
                dict_[k] = [v.__name__]
        return dict_

    @_get_field_annotation.register(list)
    def _(self, field_name) -> dict[str, Any]:
        """
        Get the annotations for multiple fields as a dictionary.

        Args:
            field_name (list): A list or tuple of field names.

        Returns:
            dict[str, Any]: A dictionary mapping field names to their
                annotations.
        """
        dict_ = {}
        for i in field_name:
            dict_.update(self._get_field_annotation(i))
        return dict_


    @_get_field_annotation.register(tuple)
    def _(self, field_name) -> dict[str, Any]:
        """
        Get the annotations for multiple fields as a dictionary.

        Args:
            field_name (tuple): A list or tuple of field names.

        Returns:
            dict[str, Any]: A dictionary mapping field names to their
                annotations.
        """
        dict_ = {}
        for i in field_name:
            dict_.update(self._get_field_annotation(i))
        return dict_


    def _field_has_attr(self, k: str, attr: str) -> bool:
        """
        Check if a field has a specific attribute.

        Args:
            k (str): The field name.
            attr (str): The attribute name.

        Returns:
            bool: True if the field has the attribute, False otherwise.
        """
        field = self.model_fields.get(k, None)
        if field is None:
            raise ValueError(f"Field {k} not found in model fields.")

        a = attr in str(field)
        if not a:
            try:
                a = (
                    self.model_fields[k].json_schema_extra[attr] is not None
                    and attr in self.model_fields[k].json_schema_extra
                )
                return a if isinstance(a, bool) else False
            except Exception:
                return False
        return a

    def to_json_str(self, *args, **kwargs) -> str:
        """
        Convert the component to a JSON string.

        Returns:
            str: The JSON string representation of the component.
        """
        return self.model_dump_json(*args, by_alias=True, **kwargs)

    def to_dict(self, *args, **kwargs) -> dict[str, Any]:
        """
        Convert the component to a dictionary.

        Returns:
            dict[str, Any]: The dictionary representation of the component.
        """
        return self.model_dump(*args, by_alias=True, **kwargs)

    def to_xml(self) -> str:
        """
        Convert the component to an XML string.

        Returns:
            str: The XML string representation of the component.
        """
        import xml.etree.ElementTree as ET

        root = ET.Element(self.__class__.__name__)

        def convert(dict_obj, parent):
            for key, val in dict_obj.items():
                if isinstance(val, dict):
                    element = ET.SubElement(parent, key)
                    convert(val, element)
                else:
                    element = ET.SubElement(parent, key)
                    element.text = str(val)

        convert(self.to_dict(), root)
        return ET.tostring(root, encoding="unicode")

    def to_pd_series(self, *args, pd_kwargs: dict | None = None, **kwargs) -> Series:
        """
        Convert the node to a Pandas Series.

        Returns:
            Series: The Pandas Series representation of the node.
        """
        pd_kwargs = {} if pd_kwargs is None else pd_kwargs
        dict_ = self.to_dict(*args, **kwargs)
        return Series(dict_, **pd_kwargs)


class BaseNode(BaseComponent):
    """
    Base class for creating node models.

    Attributes:
        content (Any | None): The optional content of the node.
        metadata (dict[str, Any]): Additional metadata for the node.
    """

    content: Any | None = Field(
        default=None,
        validation_alias=AliasChoices("text", "page_content", "chunk_content", "data"),
        description="The optional content of the node.",
    )

    metadata: dict[str, Any] = Field(
        default_factory=dict,
        alias="meta",
        description="Additional metadata for the node.",
    )

    @singledispatchmethod
    @classmethod
    def from_obj(cls, obj: Any, *args, **kwargs) -> T:
        """
        Create a node instance from an object.

        Args:
            obj (Any): The object to create the node from.

        Raises:
            NotImplementedError: If the object type is not supported.
        """
        raise NotImplementedError(f"Unsupported type: {type(obj)}")

    @from_obj.register(dict)
    @classmethod
    def _from_dict(cls, obj, *args, **kwargs) -> T:
        """
        Create a node instance from a dictionary.

        Args:
            obj (dict): The dictionary to create the node from.

        Returns:
            T: The created node instance.
        """
        return cls.model_validate(obj, *args, **kwargs)

    @from_obj.register(str)
    @classmethod
    def _from_str(cls, obj, *args, fuzzy_parse=False, **kwargs) -> T:
        """
        Create a node instance from a JSON string.

        Args:
            obj (str): The JSON string to create the node from.
            fuzzy_parse (bool): Whether to perform fuzzy parsing.

        Returns:
            T: The created node instance.
        """
        obj = ParseUtil.fuzzy_parse_json(obj) if fuzzy_parse else convert.to_dict(obj)
        try:
            return cls.from_obj(obj, *args, **kwargs)
        except ValidationError as e:
            raise ValueError(f"Invalid JSON for deserialization: {e}") from e

    @from_obj.register(list)
    @classmethod
    def _from_list(cls, obj, *args, **kwargs) -> list[T]:
        """
        Create a list of node instances from a list of objects.

        Args:
            obj (list): The list of objects to create nodes from.

        Returns:
            list[T]: The list of created node instances.
        """
        return [cls.from_obj(item, *args, **kwargs) for item in obj]

    @from_obj.register(Series)
    @classmethod
    def _from_pd_series(cls, obj, *args, pd_kwargs=None, **kwargs) -> T:
        """
        Create a node instance from a Pandas Series.

        Args:
            obj (Series): The Pandas Series to create the node from.

        Returns:
            T: The created node instance.
        """
        if pd_kwargs is None:
            pd_kwargs = {}
        return cls.from_obj(obj.to_dict(**pd_kwargs), *args, **kwargs)

    @from_obj.register(DataFrame)
    @classmethod
    def _from_pd_dataframe(cls, obj, *args, pd_kwargs=None, **kwargs) -> list[T]:
        """
        Create a list of node instances from a Pandas DataFrame.

        Args:
            obj (DataFrame): The Pandas DataFrame to create nodes from.

        Returns:
            list[T]: The list of created node instances.
        """
        if pd_kwargs is None:
            pd_kwargs = {}
        return [
            cls.from_obj(row, *args, **pd_kwargs, **kwargs) for _, row in obj.iterrows()
        ]

    @from_obj.register(BaseModel)
    @classmethod
    def _from_base_model(cls, obj, pydantic_kwargs=None, **kwargs) -> T:
        """
        Create a node instance from a Pydantic BaseModel.

        Args:
            obj (BaseModel): The Pydantic BaseModel to create the node from.

        Returns:
            T: The created node instance.
        """
        if pydantic_kwargs is None:
            pydantic_kwargs = {"by_alias": True}

        config_ = {}
        try:
            config_ = obj.model_dump(**pydantic_kwargs)
        except:
            config_ = obj.dict(**pydantic_kwargs)

        return cls.from_obj(config_ | kwargs)

    def meta_get(
        self, key: str, indices: list[str | int] = None, default: Any = None
    ) -> Any:
        """
        Get a value from the metadata dictionary.

        Args:
            key (str): The key to retrieve the value for.
            indices (list[str | int]): Optional list of indices for nested
                retrieval.
            default (Any): The default value to return if the key is not found.

        Returns:
            Any: The retrieved value or the default value if not found.
        """
        if indices:
            return nested.nget(self.metadata, indices, default)
        return self.metadata.get(key, default)

    def meta_change_key(self, old_key: str, new_key: str) -> bool:
        """
        Change a key in the metadata dictionary.

        Args:
            old_key (str): The old key to be changed.
            new_key (str): The new key to replace the old key.

        Returns:
            bool: True if the key was changed successfully, False otherwise.
        """
        if old_key in self.metadata:
            SysUtil.change_dict_key(self.metadata, old_key, new_key)
            return True
        return False

    def meta_insert(self, indices: str | list, value: Any, **kwargs) -> bool:
        """
        Insert a value into the metadata dictionary at the specified indices.

        Args:
            indices (str | list): The indices to insert the value at.
            value (Any): The value to be inserted.
            **kwargs: Additional keyword arguments for the `nested.ninsert`
                function.

        Returns:
            bool: True if the value was inserted successfully, False otherwise.
        """
        return nested.ninsert(self.metadata, indices, value, **kwargs)

    def meta_merge(
        self, additional_metadata: dict[str, Any], overwrite: bool = False, **kwargs
    ) -> None:
        """
        Merge additional metadata into the existing metadata dictionary.

        Args:
            additional_metadata (dict[str, Any]): The additional metadata to be
                merged.
            overwrite (bool): Whether to overwrite existing keys with the new
                values.
            **kwargs: Additional keyword arguments for the `nested.nmerge`
                function.
        """
        self.metadata = nested.nmerge(
            [self.metadata, additional_metadata], overwrite=overwrite, **kwargs
        )



from abc import ABC
from collections import deque
from lionagi.core.generic import BaseNode
from lionagi.core.generic.mail import Mail
from typing import Any


class Signal(BaseNode, ABC): ...


class Start(Signal):
    pending_outs: deque = deque()

    def trigger(self, context: Any, structure_id: str, executable_id: str):
        start_mail_content = {"context": context, "structure_id": structure_id}
        start_mail = Mail(
            sender=self.id_,
            recipient=executable_id,
            category="start",
            package=start_mail_content,
        )
        self.pending_outs.append(start_mail)



"""
This module contains the PredictTemplate class for predicting the next sentence(s) based on a given sentence.

The PredictTemplate class is a subclass of ScoredTemplate and provides functionality for predicting the next sentence(s)
using a language model. It includes fields for the input sentence, number of sentences to predict, predicted answer,
confidence score, and reason for the prediction.
"""

from lionagi.libs import func_call
from lionagi.integrations.bridge.pydantic_.pydantic_bridge import Field

from lionagi.core.form.scored_form import ScoredForm
from lionagi.core.branch.branch import Branch


class PredictTemplate(ScoredForm):
    """
    A class for predicting the next sentence(s) based on a given sentence.

    Attributes:
        template_name (str): The name of the predict template (default: "default_predict_template").
        sentence (str | list | dict): The given sentence(s) to predict.
        num_sentences (int): The number of sentences to predict.
        answer (str | list): The predicted sentence(s).
        signature (str): The signature indicating the input and output fields (default: "sentence -> answer").

    Methods:
        __init__(self, sentence=None, num_sentences=None, confidence_score=False, reason=False, **kwargs):
            Initializes a new instance of the PredictTemplate class.

        async predict(sentence=None, num_sentences=1, confidence_score=False, reason=False, retries=2,
                       delay=0.5, backoff_factor=2, default_value=None, timeout=None, branch_name=None,
                       system=None, messages=None, service=None, sender=None, llmconfig=None, tools=None,
                       datalogger=None, persist_path=None, tool_manager=None, **kwargs) -> PredictTemplate:
            Predicts the next sentence(s) based on the given sentence using a language model.
    """

    template_name: str = "default_predict"
    sentence: str | list | dict = Field(
        default_factory=str, description="the given sentence(s) to predict"
    )
    num_sentences: int = Field(
        default_factory=int, description="the number of sentences to predict"
    )
    answer: str | list = Field(
        default_factory=str, description="the predicted sentence(s) or desired output"
    )
    signature: str = "sentence -> answer"

    def __init__(
        self,
        sentence=None,
        instruction=None,
        num_sentences=1,
        confidence_score=False,
        reason=False,
        **kwargs,
    ):
        """
        Initializes a new instance of the PredictTemplate class.

        Args:
            sentence (Optional[str | list | dict]): The given sentence(s) to predict.
            num_sentences (Optional[int]): The number of sentences to predict.
            confidence_score (bool): Whether to include the confidence score in the output (default: False).
            reason (bool): Whether to include the reason for the prediction in the output (default: False).
            **kwargs: Additional keyword arguments.
        """
        super().__init__(**kwargs)

        self.sentence = sentence or ""
        self.num_sentences = num_sentences
        self.task = f"follow instruction to predict the next {self.num_sentences} sentence(s). Instruction: {instruction}."

        if reason:
            self.output_fields.append("reason")

        if confidence_score:
            self.output_fields.append("confidence_score")


async def predict(
    sentence=None,
    num_sentences=1,
    confidence_score=False,
    instruction=None,
    branch=None,
    reason=False,
    retries=2,
    delay=0.5,
    backoff_factor=2,
    default_value=None,
    timeout=None,
    branch_name=None,
    system=None,
    messages=None,
    service=None,
    sender=None,
    llmconfig=None,
    tools=None,
    datalogger=None,
    persist_path=None,
    tool_manager=None,
    **kwargs,
) -> "PredictTemplate":
    """
    Predicts the next sentence(s) based on the given sentence using a language model.

    Args:
        sentence (Optional[str | list | dict]): The given sentence(s) to predict.
        num_sentences (int): The number of sentences to predict (default: 1).
        confidence_score (bool): Whether to include the confidence score in the output (default: False).
        reason (bool): Whether to include the reason for the prediction in the output (default: False).
        retries (int): The number of retries for the API call (default: 2).
        delay (float): The initial delay between retries in seconds (default: 0.5).
        backoff_factor (float): The backoff factor for exponential delay between retries (default: 2).
        default_value (Optional[Any]): The default value to return if the API call fails (default: None).
        timeout (Optional[float]): The timeout for the API call in seconds (default: None).
        branch_name (Optional[str]): The name of the branch to use for prediction.
        system (Optional[Any]): The system configuration for the branch.
        messages (Optional[Any]): The messages to initialize the branch with.
        service (Optional[Any]): The service to use for prediction.
        sender (Optional[str]): The sender of the prediction request.
        llmconfig (Optional[Any]): The configuration for the language model.
        tools (Optional[Any]): The tools to use for prediction.
        datalogger (Optional[Any]): The data logger for the branch.
        persist_path (Optional[str]): The path to persist the branch data.
        tool_manager (Optional[Any]): The tool manager for the branch.
        **kwargs: Additional keyword arguments for the API call.

    Returns:
        PredictTemplate: The predict template with the predicted sentence(s).
    """
    branch = branch or Branch(
        name=branch_name,
        system=system,
        messages=messages,
        service=service,
        sender=sender,
        llmconfig=llmconfig,
        tools=tools,
        datalogger=datalogger,
        persist_path=persist_path,
        tool_manager=tool_manager,
    )

    predict_template = PredictTemplate(
        instruction=instruction,
        sentence=sentence,
        num_sentences=num_sentences,
        confidence_score=confidence_score,
        reason=reason,
    )

    await func_call.rcall(
        branch.chat,
        form=predict_template,
        retries=retries,
        delay=delay,
        backoff_factor=backoff_factor,
        default=default_value,
        timeout=timeout,
        **kwargs,
    )

    return predict_template



from lionagi.integrations.bridge.pydantic_.pydantic_bridge import Field

from lionagi.core.form.form import Form


class ScoredForm(Form):
    confidence_score: float | None = Field(
        -1,
        description="a numeric score between 0 to 1 formatted in num:0.2f",
    )
    reason: str | None = Field(
        default_factory=str, description="brief reason for the given output"
    )



from collections import deque
from abc import ABC, abstractmethod
from typing import Any

from pydantic import Field

from lionagi.core.generic import BaseComponent
from lionagi.core.mail.schema import BaseMail


class BaseExecutor(BaseComponent, ABC):
    pending_ins: dict = Field(
        default_factory=dict, description="The pending incoming mails."
    )
    pending_outs: deque = Field(
        default_factory=deque, description="The pending outgoing mails."
    )
    execute_stop: bool = Field(
        False, description="A flag indicating whether to stop execution."
    )
    context: dict | str | None = Field(
        None, description="The context buffer for the next instruction."
    )
    execution_responses: list = Field(
        default_factory=list, description="The list of responses."
    )
    context_log: list = Field(default_factory=list, description="The context log.")
    verbose: bool = Field(
        True, description="A flag indicating whether to provide verbose output."
    )
    execute_stop: bool = Field(
        False, description="A flag indicating whether to stop execution."
    )

    def send(self, recipient_id: str, category: str, package: Any) -> None:
        """
        Sends a mail to a recipient.

        Args:
            recipient_id (str): The ID of the recipient.
            category (str): The category of the mail.
            package (Any): The package to send in the mail.
        """
        mail = BaseMail(
            sender_id=self.id_,
            recipient_id=recipient_id,
            category=category,
            package=package,
        )
        self.pending_outs.append(mail)



import asyncio
from pydantic import Field

from lionagi.libs import AsyncUtil

from lionagi.core.mail.schema import BaseMail, MailTransfer
from lionagi.core.mail.mail_manager import MailManager
from lionagi.core.execute.base_executor import BaseExecutor
from lionagi.core.execute.branch_executor import BranchExecutor


class InstructionMapExecutor(BaseExecutor):
    branches: dict[str, BranchExecutor] = Field(
        default_factory=dict, description="The branches of the instruction mapping."
    )
    structure_id: str = Field("", description="The ID of the executable structure.")
    mail_transfer: MailTransfer = Field(
        default_factory=MailTransfer, description="The mail transfer."
    )
    branch_kwargs: dict = Field(
        default_factory=dict,
        description="The keyword arguments for the initializing the branches.",
    )
    num_end_branches: int = Field(0, description="The number of end branches.")
    mail_manager: MailManager = Field(
        default_factory=MailManager, description="The mail manager."
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.mail_manager = MailManager([self.mail_transfer])

    def transfer_ins(self):
        for key in list(self.pending_ins.keys()):
            while self.pending_ins[key]:
                mail: BaseMail = self.pending_ins[key].popleft()
                if mail.category == "start":
                    self._process_start(mail)
                elif mail.category == "node_list":
                    self._process_node_list(mail)
                elif (
                    (mail.category == "node")
                    or (mail.category == "condition")
                    or (mail.category == "end")
                ):
                    mail.sender_id = self.mail_transfer.id_
                    mail.recipient_id = mail.package["request_source"]
                    self.mail_transfer.pending_outs.append(mail)

    def transfer_outs(self):
        for key in list(self.mail_transfer.pending_ins.keys()):
            while self.mail_transfer.pending_ins[key]:
                mail: BaseMail = self.mail_transfer.pending_ins[key].popleft()
                if mail.category == "end":
                    self.num_end_branches += 1
                    if self.num_end_branches == len(
                        self.branches
                    ):  # tell when structure should stop
                        mail.sender_id = self.id_
                        mail.recipient_id = self.structure_id
                        self.pending_outs.append(mail)
                        self.execute_stop = True
                else:
                    mail.sender_id = self.id_
                    mail.recipient_id = self.structure_id
                    self.pending_outs.append(mail)

    def _process_start(self, start_mail: BaseMail):
        branch = BranchExecutor(verbose=self.verbose, **self.branch_kwargs)
        branch.context = start_mail.package["context"]
        self.branches[branch.id_] = branch
        self.mail_manager.add_sources([branch])
        self.structure_id = start_mail.package["structure_id"]
        mail = BaseMail(
            sender_id=self.id_,
            recipient_id=self.structure_id,
            category="start",
            package={"request_source": branch.id_, "package": "start"},
        )
        self.pending_outs.append(mail)

    def _process_node_list(self, nl_mail: BaseMail):
        source_branch_id = nl_mail.package["request_source"]
        node_list = nl_mail.package["package"]
        shared_context = self.branches[source_branch_id].context
        shared_context_log = self.branches[source_branch_id].context_log
        base_branch = self.branches[source_branch_id]

        first_node_mail = BaseMail(
            sender_id=self.mail_transfer.id_,
            recipient_id=source_branch_id,
            category="node",
            package={"request_source": source_branch_id, "package": node_list[0]},
        )
        self.mail_transfer.pending_outs.append(first_node_mail)

        for i in range(1, len(node_list)):
            branch = BranchExecutor(
                verbose=self.verbose,
                messages=base_branch.messages.copy(),
                service=base_branch.service,
                llmconfig=base_branch.llmconfig,
                datalogger=base_branch.datalogger,
            )
            branch.context = shared_context
            branch.context_log = shared_context_log
            self.branches[branch.id_] = branch
            self.mail_manager.add_sources([branch])
            node_mail = BaseMail(
                sender_id=self.mail_transfer.id_,
                recipient_id=branch.id_,
                category="node",
                package={"request_source": source_branch_id, "package": node_list[i]},
            )
            self.mail_transfer.pending_outs.append(node_mail)

    async def forward(self):
        self.transfer_ins()
        self.transfer_outs()
        self.mail_manager.collect_all()
        self.mail_manager.send_all()
        tasks = [
            branch.forward() for branch in self.branches.values() if branch.pending_ins
        ]
        await asyncio.gather(*tasks)
        return

    async def execute(self, refresh_time=1):
        while not self.execute_stop:
            await self.forward()
            await asyncio.sleep(refresh_time)



from typing import Any
from lionagi.integrations.bridge.pydantic_.pydantic_bridge import Field

from lionagi.core.form.scored_form import ScoredForm


class ActionRequest: ...


class ActionForm(ScoredForm):

    action_needed: bool | None = Field(
        False, description="true if actions are needed else false"
    )

    actions: list[dict | ActionRequest | Any] | None = Field(
        default_factory=list,
        description="""provide The list of action(s) to take, each action in {"function": function_name, "arguments": {param1:..., param2:..., ...}}. Leave blank if no further actions are needed, you must use provided parameters for each action, DO NOT MAKE UP KWARG NAME!!!""",
    )

    answer: str | dict | Any | None = Field(
        default_factory=str,
        description="output answer to the questions asked if further actions are not needed, leave blank if an accurate answer cannot be provided from context during this step",
    )

    signature: str = "sentence -> reason, action_needed, actions, answer"



from enum import Enum

from lionagi.libs import nested, convert
from lionagi.core.generic import Node
from lionagi.core.form.form import Form

_message_fields = ["node_id", "timestamp", "role", "sender", "recipient", "content"]


class BranchColumns(list[str], Enum):
    COLUMNS = _message_fields


class MessageField(str, Enum):
    NODE_ID = "node_id"
    TIMESTAMP = "timestamp"
    ROLE = "role"
    SENDER = "sender"
    RECIPIENT = "recipient"
    CONTENT = "content"
    METADATA = "metadata"
    RELATION = "relation"


class MessageRoleType(str, Enum):
    SYSTEM = "system"
    USER = "user"
    ASSISTANT = "assistant"


class MessageContentKey(str, Enum):
    INSTRUCTION = "instruction"
    CONTEXT = "context"
    SYSTEM = "system_info"
    ACTION_REQUEST = "action_request"
    ACTION_RESPONSE = "action_response"
    RESPONSE = "response"


class MessageType(dict, Enum):
    SYSTEM = (
        {
            MessageField.ROLE.value: MessageRoleType.SYSTEM.value,
            MessageField.SENDER.value: MessageRoleType.SYSTEM.value,
            MessageField.RECIPIENT.value: "null",
            "content_key": MessageContentKey.SYSTEM.value,
        },
    )

    INSTRUCTION = (
        {
            MessageField.ROLE.value: MessageRoleType.USER.value,
            MessageField.SENDER.value: MessageRoleType.USER.value,
            MessageField.RECIPIENT.value: "null",
            "content_key": MessageContentKey.INSTRUCTION.value,
        },
    )

    CONTEXT = (
        {
            MessageField.ROLE.value: MessageRoleType.USER.value,
            MessageField.SENDER.value: MessageRoleType.USER.value,
            MessageField.RECIPIENT.value: "null",
            "content_key": MessageContentKey.CONTEXT.value,
        },
    )

    ACTION_REQUEST = (
        {
            MessageField.ROLE.value: MessageRoleType.ASSISTANT.value,
            MessageField.SENDER.value: MessageRoleType.ASSISTANT.value,
            MessageField.RECIPIENT.value: "null",
            "content_key": MessageContentKey.ACTION_REQUEST.value,
        },
    )

    ACTION_RESPONSE = (
        {
            MessageField.ROLE.value: MessageRoleType.ASSISTANT.value,
            MessageField.SENDER.value: MessageRoleType.ASSISTANT.value,
            MessageField.RECIPIENT.value: "null",
            "content_key": MessageContentKey.ACTION_RESPONSE.value,
        },
    )

    RESPONSE = {
        MessageField.ROLE.value: MessageRoleType.ASSISTANT.value,
        MessageField.SENDER.value: MessageRoleType.ASSISTANT.value,
        MessageField.RECIPIENT.value: "null",
        "content_key": MessageContentKey.RESPONSE.value,
    }


class BaseMessage(Node):
    """
    Represents a message in a chatbot-like system, inheriting from BaseNode.

    Attributes:
        role (str | None): The role of the entity sending the message, e.g., 'user', 'system'.
        sender (str | None): The identifier of the sender of the message.
        content (Any): The actual content of the message.
    """

    role: str | None = None
    sender: str | None = None
    recipient: str | None = None

    @property
    def msg(self) -> dict:
        """
        Constructs and returns a dictionary representation of the message.

        Returns:
            A dictionary representation of the message with 'role' and 'content' keys.
        """
        return self._to_message()

    @property
    def msg_content(self) -> str | dict:
        """
        Gets the 'content' field of the message.

        Returns:
            The 'content' part of the message.
        """
        return self.msg["content"]

    def _to_message(self):
        """
        Constructs and returns a dictionary representation of the message.

        Returns:
            dict: A dictionary representation of the message with 'role' and 'content' keys.
        """
        return {"role": self.role, "content": convert.to_str(self.content)}

    def __str__(self):
        content_preview = (
            f"{str(self.content)[:75]}..."
            if self.content and len(self.content) > 75
            else str(self.content)
        )
        return f"Message(role={self.role}, sender={self.sender}, content='{content_preview}')"


class Instruction(BaseMessage):
    """
    Represents an instruction message, a specialized subclass of Message.

    This class is specifically used for creating messages that are instructions from the user,
    including any associated context. It sets the message role to 'user'.
    """

    def __init__(
        self,
        instruction: dict | list | str,
        context=None,
        sender: str | None = None,
        output_fields=None,
        recipient=None,
    ):
        super().__init__(
            role="user",
            sender=sender or "user",
            content={"instruction": instruction},
            recipient=recipient or "assistant",
        )
        if context:
            self.content.update({"context": context})

        if output_fields:
            format_ = f"""
            MUST EXACTLY Follow the following response format. NO ADDITIONAL COMMENTS ALLOWED!
            ```json
            {output_fields}
            ```
            """
            self.content.update(
                {"response_format": format_.replace("            ", "")}
            )

    @property
    def instruct(self):
        return self.content["instruction"]

    @classmethod
    def from_form(
        cls,
        form: Form,
        sender: str | None = None,
        recipient=None,
    ):
        return cls(
            instruction=form.instruction,
            context=form.instruction_context,
            output_fields=form.instruction_output_fields,
            sender=sender,
            recipient=recipient,
        )


class System(BaseMessage):
    """
    Represents a system-related message, a specialized subclass of Message.

    Designed for messages containing system information, this class sets the message role to 'system'.
    """

    def __init__(
        self, system: dict | list | str, sender: str | None = None, recipient=None
    ):
        super().__init__(
            role="system",
            sender=sender or "system",
            content={"system_info": system},
            recipient=recipient or "assistant",
        )

    @property
    def system_info(self):
        return self.content["system_info"]


class Response(BaseMessage):
    """
    Represents a response message, a specialized subclass of Message.

    Used for various types of response messages including regular responses, action requests,
    and action responses. It sets the message role to 'assistant'.

    """

    def __init__(
        self, response: dict | list | str, sender: str | None = None, recipient=None
    ) -> None:
        content_key = ""
        try:
            response = response["message"]
            if convert.strip_lower(response["content"]) == "none":
                content_ = self._handle_action_request(response)
                sender = sender or "action_request"
                content_key = content_key or "action_request"
                recipient = recipient or "action"

            else:
                try:
                    if "tool_uses" in convert.to_dict(response["content"]):
                        content_ = convert.to_dict(response["content"])["tool_uses"]
                        content_key = content_key or "action_request"
                        sender = sender or "action_request"
                        recipient = recipient or "action"

                    elif "response" in convert.to_dict(response["content"]):
                        sender = sender or "assistant"
                        content_key = content_key or "response"
                        content_ = convert.to_dict(response["content"])["response"]
                        recipient = recipient or "user"

                    elif "action_request" in convert.to_dict(response["content"]):
                        sender = sender or "action_request"
                        content_key = content_key or "action_request"
                        content_ = convert.to_dict(response["content"])[
                            "action_request"
                        ]
                        recipient = recipient or "action"

                    else:
                        content_ = response["content"]
                        content_key = content_key or "response"
                        sender = sender or "assistant"
                        recipient = recipient or "user"
                except Exception:
                    content_ = response["content"]
                    content_key = content_key or "response"
                    sender = sender or "assistant"
                    recipient = recipient or "user"

        except Exception:
            sender = sender or "action_response"
            content_ = response
            content_key = content_key or "action_response"
            recipient = recipient or "assistant"

        super().__init__(
            role="assistant",
            sender=sender,
            content={content_key: content_},
            recipient=recipient,
        )

    @staticmethod
    def _handle_action_request(response):
        """
        Processes an action request response and extracts relevant information.

        Args:
            response (dict): The response dictionary containing tool calls and other information.

        Returns:
            list: A list of dictionaries, each representing a function call with action and arguments.

        Raises:
            ValueError: If the response does not conform to the expected format for action requests.
        """
        try:
            tool_count = 0
            func_list = []
            while tool_count < len(response["tool_calls"]):
                _path = ["tool_calls", tool_count, "type"]

                if nested.nget(response, _path) == "function":
                    _path1 = ["tool_calls", tool_count, "function", "name"]
                    _path2 = ["tool_calls", tool_count, "function", "arguments"]

                    func_content = {
                        "action": f"action_{nested.nget(response, _path1)}",
                        "arguments": nested.nget(response, _path2),
                    }
                    func_list.append(func_content)
                tool_count += 1
            return func_list
        except:
            raise ValueError(
                "Response message must be one of regular response or function calling"
            )



from typing import overload

from abc import ABC
from collections import deque

from lionagi.libs import AsyncUtil, convert

from lionagi.core.generic import BaseNode, ActionNode, ActionSelection, Edge
from lionagi.core.tool import Tool
from lionagi.core.mail.schema import BaseMail
from lionagi.core.execute.base_executor import BaseExecutor

from lionagi.libs import AsyncUtil
from lionagi.core.generic import Node, ActionSelection, Edge
from lionagi.core.tool import Tool

from lionagi.core.mail.schema import BaseMail
from lionagi.core.graph.graph import Graph


class StructureExecutor(BaseExecutor, Graph):

    condition_check_result: bool | None = None

    async def check_edge_condition(self, edge: Edge, executable_id, request_source):
        if edge.condition.source_type == "structure":
            return edge.condition(self)

        elif edge.condition.source_type == "executable":
            return await self._check_executable_condition(
                edge, executable_id, request_source
            )

        else:
            raise ValueError("Invalid source_type.")

    def _process_edge_condition(self, edge_id):
        """
        Process the condition of a edge.

        Args:
            edge_id (str): The ID of the edge.
        """
        for key in list(self.pending_ins.keys()):
            skipped_requests = deque()
            while self.pending_ins[key]:
                mail: BaseMail = self.pending_ins[key].popleft()
                if (
                    mail.category == "condition"
                    and mail.package["package"]["edge_id"] == edge_id
                ):
                    self.condition_check_result = mail.package["package"][
                        "check_result"
                    ]
                else:
                    skipped_requests.append(mail)
            self.pending_ins[key] = skipped_requests

    async def _check_executable_condition(
        self, edge: Edge, executable_id, request_source
    ):
        self.send(
            recipient_id=executable_id,
            category="condition",
            package={"request_source": request_source, "package": edge},
        )
        while self.condition_check_result is None:
            await AsyncUtil.sleep(0.1)
            self._process_edge_condition(edge.id_)
            continue
        check_result = self.condition_check_result
        self.condition_check_result = None
        return check_result

    async def _handle_node_id(self, mail: BaseMail):
        if mail.package["package"] not in self.internal_nodes:
            raise ValueError(
                f"Node {mail.package} does not exist in the structure {self.id_}"
            )
        return await self._next_node(
            self.internal_nodes[mail.package["package"]],
            mail.sender_id,
            mail.package["request_source"],
        )

    async def _handle_node(self, mail: BaseMail):
        if not self.node_exist(mail.package["package"]):
            raise ValueError(
                f"Node {mail.package} does not exist in the structure {self.id_}"
            )
        return await self._next_node(
            mail.package["package"], mail.sender_id, mail.package["request_source"]
        )

    async def _handle_mail(self, mail: BaseMail):

        if mail.category == "start":
            return self.get_heads()

        elif mail.category == "end":
            self.execute_stop = True
            return None

        elif mail.category == "node_id":
            try:
                return await self._handle_node_id(mail)
            except Exception as e:
                raise ValueError(f"Error handling node id: {e}") from e

        elif mail.category == "node" and isinstance(mail.package["package"], BaseNode):
            try:
                return await self._handle_node(mail)
            except Exception as e:
                raise ValueError(f"Error handling node: {e}") from e

        else:
            raise ValueError(f"Invalid mail type for structure")

    async def _next_node(self, current_node: BaseNode, executable_id, request_source):
        """
        Get the next step nodes based on the current node.

        Args:
            current_node (Node): The current node.
            executable_id (str): The ID of the executable.

        Returns:
            list[Node]: The next step nodes.
        """
        next_nodes = []
        next_edges: dict[Edge] = self.get_node_edges(current_node, node_as="out")
        for edge in convert.to_list(list(next_edges.values())):
            if edge.bundle:
                continue
            if edge.condition:
                check = await self.check_edge_condition(
                    edge, executable_id, request_source
                )
                if not check:
                    continue
            node = self.internal_nodes[edge.tail]
            further_edges: dict[Edge] = self.get_node_edges(node, node_as="out")
            bundled_nodes = deque()
            for f_edge in convert.to_list(list(further_edges.values())):
                if f_edge.bundle:
                    bundled_nodes.append(self.internal_nodes[f_edge.tail])
            if bundled_nodes:
                node = self.parse_bundled_to_action(node, bundled_nodes)
            next_nodes.append(node)
        return next_nodes

    def _send_mail(self, next_nodes: list | None, mail: BaseMail):
        if not next_nodes:  # tail
            self.send(
                recipient_id=mail.sender_id,
                category="end",
                package={
                    "request_source": mail.package["request_source"],
                    "package": "end",
                },
            )
        else:
            if len(next_nodes) == 1:
                self.send(
                    recipient_id=mail.sender_id,
                    category="node",
                    package={
                        "request_source": mail.package["request_source"],
                        "package": next_nodes[0],
                    },
                )
            else:
                self.send(
                    recipient_id=mail.sender_id,
                    category="node_list",
                    package={
                        "request_source": mail.package["request_source"],
                        "package": next_nodes,
                    },
                )

    @staticmethod
    def parse_bundled_to_action(instruction: Node, bundled_nodes: deque):
        action_node = ActionNode(instruction=instruction)
        while bundled_nodes:
            node = bundled_nodes.popleft()
            if isinstance(node, ActionSelection):
                action_node.action = node.action
                action_node.action_kwargs = node.action_kwargs
            elif isinstance(node, Tool):
                action_node.tools.append(node)
            else:
                raise ValueError("Invalid bundles nodes")
        return action_node

    async def forward(self) -> None:
        """
        Process the pending incoming mails and perform the corresponding actions.
        """
        for key in list(self.pending_ins.keys()):
            while self.pending_ins[key]:
                mail: BaseMail = self.pending_ins[key].popleft()
                try:
                    if mail == "end":
                        self.execute_stop = True
                        return
                    next_nodes = await self._handle_mail(mail)
                    self._send_mail(next_nodes, mail)
                except Exception as e:
                    raise ValueError(f"Error handling mail: {e}") from e

    async def execute(self, refresh_time=1):
        if not self.acyclic:
            raise ValueError("Structure is not acyclic")

        while not self.execute_stop:
            await self.forward()
            await AsyncUtil.sleep(refresh_time)



"""
Module for representing conditions and edges between nodes in a graph structure.

This module provides the base for creating and managing edges that connect nodes
within a graph. It includes support for conditional edges, allowing the dynamic
evaluation of connections based on custom logic.
"""

from typing import Any
from pydantic import Field, field_validator
from lionagi.core.generic.component import BaseComponent, BaseNode
from lionagi.core.generic.condition import Condition


class Edge(BaseComponent):
    """
    Represents an edge between two nodes, potentially with a condition.

    Attributes:
        head (str): The identifier of the head node of the edge.
        tail (str): The identifier of the tail node of the edge.
        condition (Optional[Condition]): An optional condition that must be met
            for the edge to be considered active.
        label (Optional[str]): An optional label for the edge.

    Methods:
        check_condition: Evaluates if the condition associated with the edge is met.
    """

    head: str = Field(
        title="Head",
        description="The identifier of the head node of the edge.",
    )
    tail: str = Field(
        title="Tail",
        description="The identifier of the tail node of the edge.",
    )
    condition: Condition | None = Field(
        default=None,
        description="An optional condition that must be met for the edge to be considered active.",
    )
    label: str | None = Field(
        default=None,
        description="An optional label for the edge.",
    )
    bundle: bool = Field(
        default=False,
        description="A flag indicating if the edge is bundled.",
    )

    @field_validator("head", "tail", mode="before")
    def _validate_head_tail(cls, value):
        """
        Validates the head and tail fields to ensure they are valid node identifiers.

        Args:
            value: The value of the field being validated.
            values: A dictionary of all other values on the model.
            field: The model field being validated.

        Returns:
            The validated value, ensuring it is a valid identifier.

        Raises:
            ValueError: If the validation fails.
        """
        if isinstance(value, BaseNode):
            return value.id_
        return value

    def check_condition(self, obj: dict[str, Any]) -> bool:
        """
        Evaluates if the condition associated with the edge is met.

        Args:
            obj (dict[str, Any]): The context object used for condition evaluation.

        Returns:
            bool: True if the condition is met, False otherwise.

        Raises:
            ValueError: If the condition is not set.
        """
        if not self.condition:
            raise ValueError("The condition for the edge is not set.")
        return self.condition(obj)

    def __str__(self) -> str:
        """
        Returns a simple string representation of the Relationship.
        """

        return (
            f"Edge (id_={self.id_}, from={self.head}, to={self.tail}, "
            f"label={self.label})"
        )

    def __repr__(self) -> str:
        """
        Returns a detailed string representation of the Relationship.

        Examples:
            >>> edge = Relationship(source_node_id="node1", target_node_id="node2")
            >>> repr(edge)
            'Relationship(id_=None, from=node1, to=node2, content=None, metadata=None, label=None)'
        """
        return (
            f"Edge(id_={self.id_}, from={self.head}, to={self.tail}, "
            f"label={self.label})"
        )



from .predict import predict
from .select import select
from .score import score
from .react import react
from .vote import vote
from .plan import plan
from .cot import chain_of_thoughts, chain_of_react


__all__ = [
    "predict",
    "select",
    "score",
    "vote",
    "react",
    "plan",
    "chain_of_thoughts",
    "chain_of_react",
]



import contextlib
from lionagi.libs import convert, AsyncUtil, ParseUtil
from lionagi.core.generic import ActionNode, Edge
from lionagi.core.mail.schema import BaseMail
from lionagi.core.messages.schema import System, Instruction

from lionagi.core.branch.branch import Branch
from lionagi.core.execute.base_executor import BaseExecutor


class BranchExecutor(Branch, BaseExecutor):

    async def forward(self) -> None:
        for key in list(self.pending_ins.keys()):
            while self.pending_ins[key]:
                mail = self.pending_ins[key].popleft()
                if mail.category == "start":
                    self._process_start(mail)
                elif mail.category == "node":
                    await self._process_node(mail)
                elif mail.category == "node_list":
                    self._process_node_list(mail)
                elif mail.category == "condition":
                    self._process_condition(mail)
                elif mail.category == "end":
                    self._process_end(mail)

    async def execute(self, refresh_time=1) -> None:
        while not self.execute_stop:
            await self.forward()
            await AsyncUtil.sleep(refresh_time)

    async def _process_node(self, mail: BaseMail):
        if isinstance(mail.package["package"], System):
            self._system_process(mail.package["package"], verbose=self.verbose)
            self.send(
                mail.sender_id,
                "node_id",
                {"request_source": self.id_, "package": mail.package["package"].id_},
            )

        elif isinstance(mail.package["package"], Instruction):
            await self._instruction_process(
                mail.package["package"], verbose=self.verbose
            )
            self.send(
                mail.sender_id,
                "node_id",
                {"request_source": self.id_, "package": mail.package["package"].id_},
            )

        elif isinstance(mail.package["package"], ActionNode):
            await self._action_process(mail.package["package"], verbose=self.verbose)
            self.send(
                mail.sender_id,
                "node_id",
                {
                    "request_source": self.id_,
                    "package": mail.package["package"].instruction.id_,
                },
            )
        else:
            try:
                await self._agent_process(mail.package["package"], verbose=self.verbose)
                self.send(
                    mail.sender_id,
                    "node_id",
                    {
                        "request_source": self.id_,
                        "package": mail.package["package"].id_,
                    },
                )
            except:
                raise ValueError(f"Invalid mail to process. Mail:{mail}")

    def _process_node_list(self, mail: BaseMail):
        self.send(mail.sender_id, "end", {"request_source": self.id_, "package": "end"})
        self.execute_stop = True
        raise ValueError("Multiple path selection is currently not supported")

    def _process_condition(self, mail: BaseMail):
        relationship: Edge = mail.package["package"]
        check_result = relationship.condition(self)
        back_mail = {
            "from": self.id_,
            "edge_id": mail.package["package"].id_,
            "check_result": check_result,
        }
        self.send(
            mail.sender_id,
            "condition",
            {"request_source": self.id_, "package": back_mail},
        )

    def _system_process(self, system: System, verbose=True, context_verbose=False):
        from lionagi.libs import SysUtil

        SysUtil.check_import("IPython")
        from IPython.display import Markdown, display

        if verbose:
            print(f"------------------Welcome: {system.sender}--------------------")
            with contextlib.suppress(Exception):
                system.content = ParseUtil.fuzzy_parse_json(system.content)
            display(Markdown(f"system: {convert.to_str(system.system_info)}"))
            if self.context and context_verbose:
                display(Markdown(f"context: {convert.to_str(self.context)}"))

        self.add_message(system=system)

    async def _instruction_process(
        self, instruction: Instruction, verbose=True, **kwargs
    ):
        from lionagi.libs import SysUtil

        SysUtil.check_import("IPython")
        from IPython.display import Markdown, display

        if verbose:
            with contextlib.suppress(Exception):
                instruction.content = ParseUtil.fuzzy_parse_json(instruction.content)
            display(
                Markdown(
                    f"{instruction.sender}: {convert.to_str(instruction.instruct)}"
                )
            )

        if self.context:
            instruction.content.update({"context": self.context})
            self.context_log.append(self.context)
            self.context = None

        result = await self.chat(instruction, **kwargs)
        with contextlib.suppress(Exception):
            result = ParseUtil.fuzzy_parse_json(result)
            if "response" in result.keys():
                result = result["response"]
        if verbose and len(self.assistant_responses) != 0:
            display(
                Markdown(
                    f"{self.last_assistant_response.sender}: {convert.to_str(result)}"
                )
            )
            print("-----------------------------------------------------")

        self.execution_responses.append(result)

    async def _action_process(self, action: ActionNode, verbose=True):
        from lionagi.libs import SysUtil

        SysUtil.check_import("IPython")
        from IPython.display import Markdown, display

        try:
            func = getattr(self, action.action)
        except:
            raise ValueError(f"{action.action} is not a valid action")

        if verbose:
            display(
                Markdown(
                    f"{action.instruction.sender}: {convert.to_str(action.instruction.instruct)}"
                )
            )

        if action.tools:
            self.register_tools(action.tools)
        if self.context:
            result = await func(
                action.instruction.content["instruction"],
                context=self.context,
                tools=action.tools,
                **action.action_kwargs,
            )
            self.context = None
        else:
            result = await func(
                action.instruction.content, tools=action.tools, **action.action_kwargs
            )

        if verbose and len(self.assistant_responses) != 0:
            display(
                Markdown(
                    f"{self.last_assistant_response.sender}: {convert.to_str(result)}"
                )
            )
            print("-----------------------------------------------------")

        self.execution_responses.append(result)

    async def _agent_process(self, agent, verbose=True):
        """
        Processes an agent.

        Args:
            agent: The agent to process.
            verbose (bool): A flag indicating whether to provide verbose output (default: True).
        """
        context = self.responses
        if verbose:
            print("*****************************************************")
        result = await agent.execute(context)

        if verbose:
            print("*****************************************************")

        self.context = result
        self.responses.append(result)

    def _process_start(self, mail):
        """
        Processes a start mail.

        Args:
            mail (BaseMail): The start mail to process.
        """
        start_mail_content = mail.package
        self.context = start_mail_content["context"]
        self.send(
            start_mail_content["structure_id"],
            "start",
            {"request_source": self.id_, "package": "start"},
        )

    def _process_end(self, mail: BaseMail):
        """
        Processes an end mail.

        Args:
            mail (BaseMail): The end mail to process.
        """
        self.execute_stop = True
        self.send(mail.sender_id, "end", {"request_source": self.id_, "package": "end"})



from .component import BaseComponent, BaseNode
from .condition import Condition
from .data_logger import DataLogger, DLog
from .signal import Signal, Start
from .mail import Mail, Package
from .mailbox import MailBox
from .edge import Edge
from .relation import Relations
from .transfer import Transfer
from .work import Work, Worker
from .node import Node
from .structure import BaseStructure
from .action import ActionNode, ActionSelection


__all__ = [
    "BaseComponent",
    "BaseNode",
    "BaseStructure",
    "BaseWork",
    "Condition",
    "Edge",
    "Mail",
    "MailBox",
    "Start",
    "Package",
    "Relations",
    "Signal",
    "Transfer",
    "Node",
    "Work",
    "Worker",
    "ActionNode",
    "ActionSelection",
    "DataLogger",
    "DLog",
]



from typing import Union, Callable

from lionagi.libs import func_call
from lionagi.core.generic import Node
from ..bridge.langchain_.langchain_bridge import LangchainBridge
from ..bridge.llamaindex_.llama_index_bridge import LlamaIndexBridge


from ..loader.load_util import ChunkerType, file_to_chunks, _datanode_parser


def datanodes_convert(documents, chunker_type):

    for i in range(len(documents)):
        if type(documents[i]) == Node:
            if chunker_type == ChunkerType.LLAMAINDEX:
                documents[i] = documents[i].to_llama_index()
            elif chunker_type == ChunkerType.LANGCHAIN:
                documents[i] = documents[i].to_langchain()
    return documents


def text_chunker(documents, args, kwargs):

    def chunk_node(node):
        chunks = file_to_chunks(node.to_dict(), *args, **kwargs)
        func_call.lcall(chunks, lambda chunk: chunk.pop("node_id"))
        return [Node.from_obj({**chunk}) for chunk in chunks]

    return [chunk_node(doc) for doc in documents]


def chunk(
    documents,
    chunker,
    chunker_type=ChunkerType.PLAIN,
    chunker_args=None,
    chunker_kwargs=None,
    chunking_kwargs=None,
    documents_convert_func=None,
    to_datanode: bool | Callable = True,
):

    if chunker_args is None:
        chunker_args = []
    if chunker_kwargs is None:
        chunker_kwargs = {}
    if chunking_kwargs is None:
        chunking_kwargs = {}

    if chunker_type == ChunkerType.PLAIN:
        return chunk_funcs[ChunkerType.PLAIN](
            documents, chunker, chunker_args, chunker_kwargs
        )

    elif chunker_type == ChunkerType.LANGCHAIN:
        return chunk_funcs[ChunkerType.LANGCHAIN](
            documents,
            documents_convert_func,
            chunker,
            chunker_args,
            chunker_kwargs,
            to_datanode,
        )

    elif chunker_type == ChunkerType.LLAMAINDEX:
        return chunk_funcs[ChunkerType.LLAMAINDEX](
            documents,
            documents_convert_func,
            chunker,
            chunker_args,
            chunker_kwargs,
            to_datanode,
        )

    elif chunker_type == ChunkerType.SELFDEFINED:
        return chunk_funcs[ChunkerType.SELFDEFINED](
            documents,
            chunker,
            chunker_args,
            chunker_kwargs,
            chunking_kwargs,
            to_datanode,
        )

    else:
        raise ValueError(
            f"{chunker_type} is not supported. Please choose from {list(ChunkerType)}"
        )


def _self_defined_chunker(
    documents,
    chunker,
    chunker_args,
    chunker_kwargs,
    chunking_kwargs,
    to_datanode: bool | Callable,
):
    try:
        splitter = chunker(*chunker_args, **chunker_kwargs)
        nodes = splitter.split(documents, **chunking_kwargs)
    except Exception as e:
        raise ValueError(
            f"Self defined chunker {chunker} is not valid. Error: {e}"
        ) from e

    if isinstance(to_datanode, bool) and to_datanode is True:
        raise ValueError("Please define a valid parser to Node.")
    elif isinstance(to_datanode, Callable):
        nodes = _datanode_parser(nodes, to_datanode)
    return nodes


def _llama_index_chunker(
    documents,
    documents_convert_func,
    chunker,
    chunker_args,
    chunker_kwargs,
    to_datanode: bool | Callable,
):
    if documents_convert_func:
        documents = documents_convert_func(documents, "llama_index")
    nodes = LlamaIndexBridge.llama_index_parse_node(
        documents, chunker, chunker_args, chunker_kwargs
    )

    if isinstance(to_datanode, bool) and to_datanode is True:
        nodes = [Node.from_llama_index(i) for i in nodes]
    elif isinstance(to_datanode, Callable):
        nodes = _datanode_parser(nodes, to_datanode)
    return nodes


def _langchain_chunker(
    documents,
    documents_convert_func,
    chunker,
    chunker_args,
    chunker_kwargs,
    to_datanode: bool | Callable,
):
    if documents_convert_func:
        documents = documents_convert_func(documents, "langchain")
    nodes = LangchainBridge.langchain_text_splitter(
        documents, chunker, chunker_args, chunker_kwargs
    )
    if isinstance(to_datanode, bool) and to_datanode is True:
        if isinstance(documents, str):
            nodes = [Node(content=i) for i in nodes]
        else:
            nodes = [Node.from_langchain(i) for i in nodes]
    elif isinstance(to_datanode, Callable):
        nodes = _datanode_parser(nodes, to_datanode)
    return nodes


def _plain_chunker(documents, chunker, chunker_args, chunker_kwargs):
    try:
        if chunker == "text_chunker":
            chunker = text_chunker
        return chunker(documents, chunker_args, chunker_kwargs)
    except Exception as e:
        raise ValueError(
            f"Reader {chunker} is currently not supported. Error: {e}"
        ) from e


chunk_funcs = {
    ChunkerType.PLAIN: _plain_chunker,
    ChunkerType.LANGCHAIN: _langchain_chunker,
    ChunkerType.LLAMAINDEX: _llama_index_chunker,
    ChunkerType.SELFDEFINED: _self_defined_chunker,
}



from lionagi.libs import func_call, AsyncUtil

from lionagi.integrations.bridge.pydantic_.pydantic_bridge import Field
from lionagi.core.form.action_form import ActionForm
from lionagi.core.branch.branch import Branch
from lionagi.core.direct.utils import _process_tools


class ReactTemplate(ActionForm):
    template_name: str = "default_react"
    sentence: str | list | dict | None = Field(
        default_factory=str,
        description="the given sentence(s) to reason and take actions on",
    )

    def __init__(
        self,
        sentence=None,
        instruction=None,
        confidence_score=False,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.sentence = sentence
        self.task = f"Think step by step. Perform reasoning and prepare actions with given tools only.Instruction: {instruction}. Absolutely DO NOT MAKE UP FUNCTIONS !!!"

        if confidence_score:
            self.output_fields.append("confidence_score")


async def _react(
    sentence=None,
    *,
    instruction=None,
    branch=None,
    confidence_score=False,
    retries=2,
    delay=0.5,
    backoff_factor=2,
    default_value=None,
    timeout=None,
    branch_name=None,
    system=None,
    messages=None,
    service=None,
    sender=None,
    llmconfig=None,
    tools=None,
    datalogger=None,
    persist_path=None,
    tool_manager=None,
    return_branch=False,
    **kwargs,
):

    if "temperature" not in kwargs:
        kwargs["temperature"] = 0.1

    instruction = instruction or ""

    if branch and tools:
        _process_tools(tools, branch)

    branch = branch or Branch(
        name=branch_name,
        system=system,
        messages=messages,
        service=service,
        sender=sender,
        llmconfig=llmconfig,
        tools=tools,
        datalogger=datalogger,
        persist_path=persist_path,
        tool_manager=tool_manager,
    )

    _template = ReactTemplate(
        sentence=sentence,
        instruction=instruction,
        confidence_score=confidence_score,
    )

    await func_call.rcall(
        branch.chat,
        form=_template,
        retries=retries,
        delay=delay,
        backoff_factor=backoff_factor,
        default=default_value,
        timeout=timeout,
        **kwargs,
    )

    if _template.action_needed:
        actions = _template.actions
        tasks = [branch.tool_manager.invoke(i.values()) for i in actions]
        results = await AsyncUtil.execute_tasks(*tasks)

        a = []
        for idx, item in enumerate(actions):
            res = {
                "function": item["function"],
                "arguments": item["arguments"],
                "output": results[idx],
            }
            branch.add_message(response=res)
            a.append(res)

        _template.__setattr__("action_response", a)

    return (_template, branch) if return_branch else _template


async def react(
    sentence=None,
    *,
    instruction=None,
    num_instances=1,
    branch=None,
    confidence_score=False,
    retries=2,
    delay=0.5,
    backoff_factor=2,
    default_value=None,
    timeout=None,
    branch_name=None,
    system=None,
    messages=None,
    service=None,
    sender=None,
    llmconfig=None,
    tools=None,
    datalogger=None,
    persist_path=None,
    tool_manager=None,
    return_branch=False,
    **kwargs,
):

    async def _inner(i=0):
        return await _react(
            sentence=sentence,
            instruction=instruction,
            num_instances=num_instances,
            branch=branch,
            confidence_score=confidence_score,
            retries=retries,
            delay=delay,
            backoff_factor=backoff_factor,
            default_value=default_value,
            timeout=timeout,
            branch_name=branch_name,
            system=system,
            messages=messages,
            service=service,
            sender=sender,
            llmconfig=llmconfig,
            tools=tools,
            datalogger=datalogger,
            persist_path=persist_path,
            tool_manager=tool_manager,
            return_branch=return_branch,
            **kwargs,
        )

    if num_instances == 1:
        return await _inner()

    elif num_instances > 1:
        return await func_call.alcall(range(num_instances), _inner)



"""
A module for representing relationships between nodes in a graph structure, 
encapsulating incoming and outgoing edges.
"""

from pydantic import Field
from pydantic.dataclasses import dataclass
from lionagi.libs import convert
from lionagi.core.generic.edge import Edge


@dataclass
class Relations:
    """
    Represents the relationships of a node via its incoming and outgoing edges.

    This class stores edges in two dictionaries: `preceding` for outgoing edges
    and `succeeding` for incoming edges. It provides properties to access all
    edges together and to get a unique set of all connected node IDs.

    Attributes:
        preceding (dict[str, Edge]): A dictionary of outgoing edges from the
            node, with the edge ID as the key and the `Edge` object as the
            value. Represents edges leading from this node to other nodes.
        succeeding (dict[str, Edge]): A dictionary of incoming edges to the
            node, with the edge ID as the key and the `Edge` object as the
            value. Represents edges from other nodes leading to this node.
    """

    points_to: dict[str, Edge] = Field(
        title="Outgoing edges",
        default_factory=dict,
        description="The Outgoing edges of the node, reads self precedes other, \
            {edge_id: Edge}",
    )

    pointed_by: dict[str, Edge] = Field(
        title="Incoming edges",
        default_factory=dict,
        description="The Incoming edges of the node, reads self succeeds other, \
            {edge_id: Edge}",
    )

    @property
    def all_edges(self) -> dict[str, Edge]:
        """
        Combines and returns all incoming and outgoing edges of the node.

        Returns:
            dict[str, Edge]: A dictionary of all edges connected to the node,
                including both preceding (outgoing) and succeeding (incoming)
                edges, indexed by edge IDs.
        """
        return {**self.points_to, **self.pointed_by}

    @property
    def all_nodes(self) -> set[str]:
        """
        Extracts and returns a unique set of all node IDs connected to this
        node through its edges.

        It processes both heads and tails of each edge in `all_edges`, flattens
        the list to a one-dimensional list, and then converts it to a set to
        ensure uniqueness.

        Returns:
            set[str]: A set of unique node IDs connected to this node, derived
                from both incoming and outgoing edges.
        """
        return set(convert.to_list([[i.head, i.tail] for i in self.all_edges.values()]))



from abc import ABC
from typing import Any
from lionagi.core.form.field_validator import validation_funcs


from lionagi.libs import convert, func_call

non_prompt_words = [
    "id_",
    "node_id",
    "meta",
    "metadata",
    "timestamp",
    "content",
    "signature",
    "task",
    "template_name",
    "version",
    "description",
    "in_validation_kwargs",
    "out_validation_kwargs",
    "fix_input",
    "fix_output",
    "input_fields",
    "output_fields",
    "choices",
    "prompt_fields",
    "prompt_fields_annotation",
    "instruction_context",
    "instruction",
    "instruction_output_fields",
    "in_",
    "out",
    "process",
    "_validate_field",
    "_process_input",
    "_process_response",
    "_validate_field_choices",
    "_validate_input_choices",
    "_validate_output_choices",
]


class PromptTemplateMixin(ABC):
    def _validate_field_choices(self, fields, fix_: bool = False):
        if len(self.choices) >= 1:
            for k, choices in self.choices.items():
                if k in fields and not self._validate_field(
                    k, getattr(self, k), choices, fix_
                ):
                    raise ValueError(
                        f"Invalid choice for field {k}: {getattr(self, k)} is not in {choices}"
                    )

    def _validate_input_choices(self):
        return self._validate_field_choices(self.input_fields, self.fix_input)

    def _validate_output_choices(self):
        return self._validate_field_choices(self.output_fields, self.fix_output)

    def _validate_field(self, k, v, choices=None, keys=None, fix_=False, **kwargs):

        str_ = self._prompt_fields_annotation[k]

        if keys:
            v_ = validation_funcs["dict"](v, keys=keys, fix_=fix_, **kwargs)
            setattr(self, k, v_)
            return True

        if choices:
            v_ = validation_funcs["enum"](v, choices=choices, fix_=fix_, **kwargs)
            if v_ not in choices:
                raise ValueError(f"{v} is not in chocies {choices}")
            setattr(self, k, v_)
            return True

        if "lionagi.core.prompt.action_template.actionrequest" in str_:
            self.__setattr__(k, validation_funcs["action"](v))
            return True

        if "bool" in str_ and "str" not in str_:
            self.__setattr__(k, validation_funcs["bool"](v, fix_=fix_, **kwargs))
            return True

        if any(i in str_ for i in ["int", "float", "number"]) and "str" not in str_:
            self.__setattr__(k, validation_funcs["number"](v, fix_=fix_, **kwargs))
            return True

        elif "str" in str_:
            self.__setattr__(k, validation_funcs["str"](v, fix_=fix_, **kwargs))
            return True

        return False

    def _process_input(self, fix_=False):
        kwargs = self.in_validation_kwargs.copy()
        for k, v in self.in_.items():
            if k not in kwargs:
                kwargs = {k: {}}

            if self._field_has_choices(k):
                self.choices[k] = self.model_fields[k].json_schema_extra["choices"]
                if self._validate_field(
                    k, v, choices=self.choices[k], fix_=fix_, **kwargs[k]
                ):
                    continue
                else:
                    raise ValueError(f"{k} has no choices")

            elif self._validate_field(k, v, fix_=fix_, **kwargs[k]):
                continue
            else:
                raise ValueError(f"failed to validate field {k}")

    def _get_field_attr(self, k, attr) -> Any:
        if not self._field_has_attr(k, attr):
            raise ValueError(f"field {k} has no attribute {attr}")
        field = self.model_fields[k]
        a = getattr(field, attr)
        if not a:
            try:
                a = field.json_schema_extra[attr]
                return a
            except Exception:
                return None
        return a

    def _field_has_attr(self, k, attr):
        a = False

        field = self.model_fields[k]
        a = hasattr(field, attr)

        if not a:
            try:
                a = (
                    self.model_fields[k].json_schema_extra[attr] is not None
                    and attr in self.model_fields[k].json_schema_extra
                )
                return a if isinstance(a, bool) else False
            except Exception:
                return False

    def _field_has_keys(self, k):
        return self._field_has_attr(k, "keys")

    def _field_has_choices(self, k):
        return self._field_has_attr(k, "choices")

    def _process_choices(self, k, v, fix_=False, kwargs=None):
        choices = self._get_field_attr(k, "choices")

        if self._validate_field(k, v, choices=choices, fix_=fix_, **kwargs):
            self.choices[k] = choices
            return True
        else:
            raise ValueError(f"{k} has no choices")

    def _process_keys(self, k, v, fix_=False, kwargs=None):
        keys = self._get_field_attr(k, "keys")
        if self._validate_field(k, v, keys=keys, fix_=fix_, **kwargs):
            return True
        else:
            raise ValueError(f"{k} has no keys")

    def _process_response(self, out_, fix_=True):
        kwargs = self.out_validation_kwargs.copy()
        for k, v in out_.items():
            if k not in kwargs:
                kwargs = {k: {}}

            if self._field_has_choices(k):
                try:
                    return self._process_choices(k, v, fix_=fix_, kwargs=kwargs[k])
                except Exception as e:
                    raise ValueError(
                        f"failed to process field {k} with value {v}"
                    ) from e

            elif self._field_has_keys(k):
                try:
                    return self._process_keys(k, v, fix_=fix_, kwargs=kwargs[k])
                except Exception as e:
                    raise ValueError(
                        f"failed to process field {k} with value {v}"
                    ) from e

            elif self._validate_field(k, v, fix_=fix_, **kwargs[k]):
                continue

            else:
                raise ValueError(f"failed to validate field {k} with value {v}")

    @staticmethod
    def _get_input_output_fields(str_: str) -> Any:
        inputs, outputs = str_.split("->")

        input_fields = [convert.strip_lower(i) for i in inputs.split(",")]
        output_fields = [convert.strip_lower(o) for o in outputs.split(",")]

        return input_fields, output_fields

    @property
    def _prompt_fields_annotation(self):
        dict_ = {i: self.model_fields[i].annotation for i in self.prompt_fields}
        for k, v in dict_.items():
            if "|" in str(v):
                v = str(v)
                v = v.split("|")
                dict_[k] = func_call.lcall(v, convert.strip_lower)
            else:
                dict_[k] = [v.__name__]

        return dict_



"""
This module defines classes for representing mail packages and mailboxes
in a messaging system.

The module includes the following classes:
- MailPackageCategory: An enumeration of categories for mail packages.
- Mail: Represents a mail message sent from one component to another.
- MailBox: Represents a mailbox that stores pending incoming and outgoing mails.
"""

from typing import Any
from enum import Enum

from pydantic import Field, field_validator

from lionagi.core.generic.component import BaseComponent


class MailPackageCategory(str, Enum):
    """
    Defines categories for mail packages in a messaging system.

    Attributes:
        MESSAGES: Represents general messages.
        TOOL: Represents tools.
        SERVICE: Represents services.
        MODEL: Represents models.
        NODE: Represents nodes.
        NODE_LIST: Represents a list of nodes.
        NODE_ID: Represents a node's ID.
        START: Represents a start signal or value.
        END: Represents an end signal or value.
        CONDITION: Represents a condition.
    """

    MESSAGES = "messages"
    TOOL = "tool"
    SERVICE = "service"
    MODEL = "model"
    NODE = "node"
    NODE_LIST = "node_list"
    NODE_ID = "node_id"
    START = "start"
    END = "end"
    CONDITION = "condition"


class Package(BaseComponent):
    category: MailPackageCategory = Field(
        ..., title="Category", description="The category of the mail package."
    )

    package: Any = Field(
        ..., title="Package", description="The package to send in the mail."
    )


class Mail(BaseComponent):
    """
    Represents a mail message sent from one component to another within
    the system.

    Attributes:
        sender (str): The ID of the sender node.
        recipient (str): The ID of the recipient node.
        category (MailPackageCategory): The category of the mail package.
        package (Any): The content of the mail package.
    """

    sender: str = Field(
        title="Sender",
        description="The id of the sender node.",
    )

    recipient: str = Field(
        title="Recipient",
        description="The id of the recipient node.",
    )

    packages: dict[str, Package] = Field(
        title="Packages",
        default_factory=dict,
        description="The packages to send in the mail.",
    )

    @field_validator("sender", "recipient", mode="before")
    def _validate_sender_recipient(cls, value):
        if isinstance(value, BaseComponent):
            return value.id_
        return value



from pydantic.dataclasses import dataclass
from typing import Any
from abc import ABC, abstractmethod
from pydantic import Field

from .node import Node


@dataclass
class Work:

    work_completed: bool = Field(
        default=False,
        description="A flag indicating whether the work is already completed.",
    )

    work_result: Any | None = Field(
        default=None,
        description="The result of the work.",
    )

    context: dict | str | None = Field(
        None, description="The context buffer for the next instruction."
    )


class Worker(Node, ABC):
    stopped: bool = False

    @abstractmethod
    def perform(self, *args, **kwargs):
        pass

    @abstractmethod
    def forward(self, *args, **kwargs):
        pass

    def stop(self):
        self.stopped = True
        return True



from collections import deque
from pydantic import Field
from pydantic.dataclasses import dataclass

from lionagi.core.generic.mail import Mail


@dataclass
class MailBox:

    pile: dict[str, Mail] = Field(
        default_factory=dict, description="The pile of all mails - {mail_id: Mail}"
    )

    sequence_in: dict[str, deque] = Field(
        default_factory=dict,
        description="The sequence of all incoming mails - {sender_id: deque[mail_id]}",
    )

    sequence_out: deque = Field(
        default_factory=deque,
        description="The sequence of all outgoing mails - deque[mail_id]",
    )

    def __str__(self) -> str:
        """
        Returns a string representation of the MailBox instance.

        Returns:
            str: A string describing the number of pending incoming and
                outgoing mails in the MailBox.
        """
        return (
            f"MailBox with {len(self.receieving)} pending incoming mails and "
            f"{len(self.sending)} pending outgoing mails."
        )



# plan.py

from lionagi.libs import func_call, ParseUtil
from lionagi.integrations.bridge.pydantic_.pydantic_bridge import Field
from lionagi.core.form.scored_form import ScoredForm
from lionagi.core.branch.branch import Branch


class PlanTemplate(ScoredForm):
    template_name: str = "default_plan"
    sentence: str | list | dict = Field(
        default_factory=str,
        description="the given sentence(s) or context to generate a plan for",
    )
    plan: dict | str = Field(
        default_factory=dict,
        description="the generated step by step plan, return as a dictionary following {step_n: {plan: ..., reason: ...}} format",
    )
    signature: str = "sentence -> plan"

    def __init__(
        self,
        sentence=None,
        instruction=None,
        confidence_score=False,
        reason=False,
        num_step=3,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.sentence = sentence
        self.task = f"Generate a {num_step}_step plan based on the given context. Instruction: {instruction}."

        if reason:
            self.output_fields.append("reason")

        if confidence_score:
            self.output_fields.append("confidence_score")


async def _plan(
    sentence,
    *,
    instruction=None,
    branch=None,
    confidence_score=False,
    reason=False,
    retries=2,
    delay=0.5,
    backoff_factor=2,
    default_value=None,
    timeout=None,
    branch_name=None,
    system=None,
    messages=None,
    service=None,
    sender=None,
    llmconfig=None,
    tools=None,
    datalogger=None,
    persist_path=None,
    tool_manager=None,
    return_branch=False,
    **kwargs,
):
    if "temperature" not in kwargs:
        kwargs["temperature"] = 0.1

    instruction = instruction or ""

    branch = branch or Branch(
        name=branch_name,
        system=system,
        messages=messages,
        service=service,
        sender=sender,
        llmconfig=llmconfig,
        tools=tools,
        datalogger=datalogger,
        persist_path=persist_path,
        tool_manager=tool_manager,
    )

    _template = PlanTemplate(
        sentence=sentence,
        instruction=instruction,
        confidence_score=confidence_score,
        reason=reason,
    )

    await func_call.rcall(
        branch.chat,
        form=_template,
        retries=retries,
        delay=delay,
        backoff_factor=backoff_factor,
        default=default_value,
        timeout=timeout,
        **kwargs,
    )

    _template.plan = ParseUtil.fuzzy_parse_json(_template.plan)

    return (_template, branch) if return_branch else _template


async def plan(
    sentence,
    *,
    instruction=None,
    num_instances=1,
    branch=None,
    confidence_score=False,
    reason=False,
    retries=2,
    delay=0.5,
    backoff_factor=2,
    default_value=None,
    timeout=None,
    branch_name=None,
    system=None,
    messages=None,
    service=None,
    sender=None,
    llmconfig=None,
    tools=None,
    datalogger=None,
    persist_path=None,
    tool_manager=None,
    return_branch=False,
    **kwargs,
):
    async def _inner(i=0):
        return await _plan(
            sentence=sentence,
            instruction=instruction,
            branch=branch,
            confidence_score=confidence_score,
            reason=reason,
            retries=retries,
            delay=delay,
            backoff_factor=backoff_factor,
            default_value=default_value,
            timeout=timeout,
            branch_name=branch_name,
            system=system,
            messages=messages,
            service=service,
            sender=sender,
            llmconfig=llmconfig,
            tools=tools,
            datalogger=datalogger,
            persist_path=persist_path,
            tool_manager=tool_manager,
            return_branch=return_branch,
            **kwargs,
        )

    if num_instances == 1:
        return await _inner()

    elif num_instances > 1:
        return await func_call.alcall(range(num_instances), _inner)



from collections import deque
from pydantic import Field
from pydantic.dataclasses import dataclass

from lionagi.core.generic.mail import Mail
from lionagi.core.generic.signal import Signal


@dataclass
class Transfer:

    schedule: dict[str, deque[Mail | Signal]] = Field(
        default_factory=dict,
        description="The sequence of all pending mails - {direction: deque[mail_id]}",
    )

    @property
    def is_empty(self) -> bool:
        """Returns a flag indicating whether the transfer is empty."""
        return not any(self.schedule.values())



from abc import ABC, abstractmethod
from enum import Enum
from pydantic import BaseModel, Field


class ConditionSource(str, Enum):
    """
    Enumeration for specifying the source type of a condition.

    Attributes:
        STRUCTURE: Represents a condition based on the structure.
        EXECUTABLE: Represents a condition that can be executed or evaluated.
    """

    STRUCTURE = "structure"
    EXECUTABLE = "executable"


class Condition(BaseModel, ABC):
    """
    Abstract base class for defining conditions associated with edges.

    Attributes:
        source_type (ConditionSource): Specifies the type of source for the condition.

    Methods:
        check: Abstract method that should be implemented to evaluate the condition.
    """

    source_type: str = Field(..., description="The type of source for the condition.")

    class Config:
        """Model configuration settings."""

        extra = "allow"

    @abstractmethod
    def __call__(self, *args, **kwargs) -> bool:
        """Evaluates the condition based on implemented logic.

        Returns:
            bool: The boolean result of the condition evaluation.
        """
        pass



from typing import Any, Type
from pydantic import Field
from lionagi.integrations.bridge import LlamaIndexBridge, LangchainBridge

from lionagi.core.generic.component import BaseNode
from lionagi.core.generic.condition import Condition
from lionagi.core.generic.edge import Edge
from lionagi.core.generic.relation import Relations
from lionagi.core.generic.mailbox import MailBox


class Node(BaseNode):
    """
    Represents a node with relations to other nodes.

    Attributes:
        relations (Relations): The relations of the node, managed through a
            `Relations` instance.

    Properties:
        related_nodes: A set of IDs representing nodes related to this node.
        edges: A dictionary of all edges connected to this node.
        node_relations: A dictionary categorizing preceding and succeeding
            relations to this node.
        precedessors: A list of node IDs that precede this node.
        successors: A list of node IDs that succeed this node.

    Methods:
        relate(node, self_as, condition, **kwargs): Relates this node to
            another node with an edge.
        unrelate(node, edge): Removes one or all relations between this node
            and another.
        to_llama_index(node_type, **kwargs): Serializes this node for
            LlamaIndex.
        to_langchain(**kwargs): Serializes this node for Langchain.
        from_llama_index(llama_node, **kwargs): Deserializes a node from
            LlamaIndex data.
        from_langchain(lc_doc): Deserializes a node from Langchain data.
        __str__(): String representation of the node.

    Raises:
        ValueError: When invalid parameters are provided to methods.
    """

    relations: Relations = Field(
        default_factory=Relations,
        description="The relations of the node.",
        alias="node_relations",
    )

    mailbox: MailBox = Field(
        default_factory=MailBox,
        description="The mailbox for incoming and outgoing mails.",
    )

    @property
    def related_nodes(self) -> list[str]:
        """Returns a set of node IDs related to this node, excluding itself."""
        nodes = set(self.relations.all_nodes)
        nodes.discard(self.id_)
        return list(nodes)

    @property
    def edges(self) -> dict[str, Edge]:
        """Returns a dictionary of all edges connected to this node."""
        return self.relations.all_edges

    @property
    def node_relations(self) -> dict:
        """Categorizes preceding and succeeding relations to this node."""

        points_to_nodes = {}
        for edge in self.relations.points_to.values():
            for i in self.related_nodes:
                if edge.tail == i:
                    if i in points_to_nodes:
                        points_to_nodes[i].append(edge)
                    else:
                        points_to_nodes[i] = [edge]

        pointed_by_nodes = {}
        for edge in self.relations.pointed_by.values():
            for i in self.related_nodes:
                if edge.head == i:
                    if i in pointed_by_nodes:
                        pointed_by_nodes[i].append(edge)
                    else:
                        pointed_by_nodes[i] = [edge]

        return {"points_to": points_to_nodes, "pointed_by": pointed_by_nodes}

    @property
    def precedessors(self) -> list[str]:
        """return a list of nodes id that precede this node"""
        return [k for k, v in self.node_relations["pointed_by"].items() if len(v) > 0]

    @property
    def successors(self) -> list[str]:
        """return a list of nodes id that succeed this node"""
        return [k for k, v in self.node_relations["points_to"].items() if len(v) > 0]

    def relate(
        self,
        node: "Node",
        node_as: str = "head",
        condition: Condition | None = None,
        label: str | None = None,
        bundle=False,
    ) -> None:
        """Relates this node to another node with an edge.

        Args:
            node (Node): The node to relate to.
            self_as (str): Specifies whether this node is the 'head' or 'tail'
                of the relation. Defaults to "head".
            condition (Condition | None): The condition associated with the
                edge, if any. Defaults to None.
            **kwargs: Additional keyword arguments for edge creation.

        Raises:
            ValueError: If `self_as` is not 'head' or 'tail'.
        """
        if node_as == "head":
            edge = Edge(
                head=self, tail=node, condition=condition, bundle=bundle, label=label
            )
            self.relations.points_to[edge.id_] = edge
            node.relations.pointed_by[edge.id_] = edge

        elif node_as == "tail":
            edge = Edge(
                head=node, tail=self, condition=condition, label=label, bundle=bundle
            )
            self.relations.pointed_by[edge.id_] = edge
            node.relations.points_to[edge.id_] = edge

        else:
            raise ValueError(
                f"Invalid value for self_as: {node_as}, must be 'head' or 'tail'"
            )

    def remove_edge(self, node: "Node", edge: Edge | str) -> bool:
        if node.id_ not in self.related_nodes:
            raise ValueError(f"Node {self.id_} is not related to node {node.id_}.")

        edge_id = edge.id_ if isinstance(edge, Edge) else edge

        if (
            edge_id not in self.relations.all_edges
            or edge_id not in node.relations.all_edges
        ):
            raise ValueError(
                f"Edge {edge_id} does not exist between nodes {self.id_} and "
                f"{node.id_}."
            )

        all_dicts = [
            self.relations.points_to,
            self.relations.pointed_by,
            node.relations.points_to,
            node.relations.pointed_by,
        ]
        try:
            for _dict in all_dicts:
                edge_id = edge.id_ if isinstance(edge, Edge) else edge
                _dict.pop(edge_id, None)
            return True

        except Exception as e:
            raise ValueError(
                f"Failed to remove edge between nodes {self.id_} and " f"{node.id_}."
            ) from e

    def unrelate(self, node: "Node", edge: Edge | str = "all") -> bool:
        """
        Removes one or all relations between this node and another.

        Args:
            node (Node): The node to unrelate from.
            edge (Edge | str): Specific edge or 'all' to remove all relations.
                Defaults to "all".

        Returns:
            bool: True if the operation is successful, False otherwise.

        Raises:
            ValueError: If the node is not related or the edge does not exist.
        """
        if edge == "all":
            edge = self.node_relations["points_to"].get(
                node.id_, []
            ) + self.node_relations["pointed_by"].get(node.id_, [])
        else:
            edge = [edge.id_] if isinstance(edge, Edge) else [edge]

        if len(edge) == 0:
            raise ValueError(f"Node {self.id_} is not related to node {node.id_}.")

        try:
            for edge_id in edge:
                self.remove_edge(node, edge_id)
            return True
        except Exception as e:
            raise ValueError(
                f"Failed to remove edge between nodes {self.id_} and " f"{node.id_}."
            ) from e

    def to_llama_index(self, node_type: Type | str | Any = None, **kwargs) -> Any:
        """
        Serializes this node for LlamaIndex.

        Args:
            node_type (Type | str | Any): The type of node in LlamaIndex.
                Defaults to None.
            **kwargs: Additional keyword arguments for serialization.

        Returns:
            Any: The serialized node for LlamaIndex.
        """
        return LlamaIndexBridge.to_llama_index_node(self, node_type=node_type, **kwargs)

    def to_langchain(self, **kwargs) -> Any:
        """
        Serializes this node for Langchain.

        Args:
            **kwargs: Additional keyword arguments for serialization.

        Returns:
            Any: The serialized node for Langchain.
        """
        return LangchainBridge.to_langchain_document(self, **kwargs)

    @classmethod
    def from_llama_index(cls, llama_node: Any, **kwargs) -> "Node":
        """
        Deserializes a node from LlamaIndex data.

        Args:
            llama_node (Any): The LlamaIndex node data.
            **kwargs: Additional keyword arguments for deserialization.

        Returns:
            Node: The deserialized node.
        """
        llama_dict = llama_node.to_dict(**kwargs)
        return cls.from_obj(llama_dict)

    @classmethod
    def from_langchain(cls, lc_doc: Any) -> "Node":
        """Deserializes a node from Langchain data.

        Args:
            lc_doc (Any): The Langchain document data.

        Returns:
            Node: The deserialized node.
        """
        langchain_json = lc_doc.to_json()
        langchain_dict = {"lc_id": langchain_json["id"], **langchain_json["kwargs"]}
        return cls.from_obj(langchain_dict)

    def __str__(self) -> str:
        """
        Provides a string representation of the node.

        Returns:
            str: The string representation of the node.
        """
        timestamp = f" ({self.timestamp})" if self.timestamp else ""
        if self.content:
            content_preview = (
                f"{self.content[:50]}..." if len(self.content) > 50 else self.content
            )
        else:
            content_preview = ""
        meta_preview = (
            f"{str(self.metadata)[:50]}..."
            if len(str(self.metadata)) > 50
            else str(self.metadata)
        )
        return (
            f"{self.class_name()}({self.id_}, {content_preview}, {meta_preview},"
            f"{timestamp})"
        )



from . import *

from .branch.branch import Branch
from .session.session import Session
from .generic import (
    ActionNode,
    ActionSelection,
    Condition,
)
from .agent.base_agent import BaseAgent
from .messages.schema import Instruction, System, Response
from .tool import func_to_tool


__all__ = [
    "ActionNode",
    "ActionSelection",
    "Branch",
    "Condition",
    "Session",
    "System",
    "Instruction",
    "Response",
    "BaseAgent",
    "func_to_tool",
]



from lionagi.libs import convert

from lionagi.core.direct.predict import predict
from lionagi.core.direct.plan import plan
from lionagi.core.direct.react import react

from .utils import _process_tools


async def chain_of_thoughts(
    sentence=None,
    branch=None,
    instruction=None,
    reason=False,
    confidence_score=False,
    num_steps=3,
    directive_kwargs={},
    return_branch=False,
    **kwargs,
):

    out_, outs, answer, reasons, confidence_score = "", [], "", [], 0
    if branch is not None:
        out_ = await plan(
            sentence,
            branch=branch,
            instruction=instruction,
            num_steps=num_steps,
            **kwargs,
        )
    else:
        out_, branch = await plan(
            sentence,
            instruction=instruction,
            branch=branch,
            num_steps=num_steps,
            return_branch=True,
            **kwargs,
        )

    for i in range(len(out_.plan)):
        _out = await predict(
            branch=branch,
            instruction=out_.plan[f"step_{i+1}"],
            reason=reason,
            confidence_score=confidence_score,
            **directive_kwargs,
        )
        answer += _out.answer
        if reason:
            reasons.append(_out.reason)
        if confidence_score:
            confidence_score += _out.confidence_score
        outs.append(_out)

    setattr(out_, "chain_output", outs)
    setattr(out_, "chain_answer", answer)

    if reason:
        setattr(out_, "chain_reasons", reasons)
    if confidence_score:
        setattr(out_, "chain_confidence_score", confidence_score / len(outs))

    if return_branch:
        return out_, branch

    return out_


async def chain_of_react(
    sentence=None,
    branch=None,
    instruction=None,
    num_steps=3,
    tools=None,
    directive_system=None,
    directive_kwargs={},
    return_branch=False,
    **kwargs,
):
    out_, outs, reasons, actions, action_responses = "", [], [], [], []
    if branch is not None:
        out_ = await plan(
            sentence,
            branch=branch,
            instruction=instruction,
            num_steps=num_steps,
            **kwargs,
        )
    else:
        out_, branch = await plan(
            sentence,
            instruction=instruction,
            branch=branch,
            num_steps=num_steps,
            return_branch=True,
            **kwargs,
        )

    _process_tools(tools, branch)

    for i in range(len(out_.plan)):
        _out = await react(
            branch=branch,
            system=directive_system,
            instruction=out_.plan[f"step_{i+1}"],
            **directive_kwargs,
        )
        outs.append(_out)
        reasons.append(_out.reason)
        actions.append(_out.actions)
        if _out.action_needed:
            action_responses.append(_out.action_response)

    setattr(out_, "chain_output", convert.to_list(outs))
    setattr(out_, "chain_reason", convert.to_list(reasons))
    setattr(out_, "chain_actions", convert.to_list(actions))
    setattr(out_, "chain_action_response", convert.to_list(action_responses))

    if return_branch:
        return out_, branch

    return out_



"""
This module provides classes for representing and working with tree structures.

The module includes the following classes:
- TreeLabel: An enumeration representing tree relationships (parent and child).
- TreeNode: A specialized node representing a node within a tree structure,
  extending the basic node functionality with parent-child relationships.
- Tree: Represents a tree structure, extending the graph with tree-specific
  functionalities like managing parent-child relationships.
"""

from enum import Enum
from pydantic import Field

from lionagi.core.generic import Node, Condition
from lionagi.core.graph.graph import Graph


class TreeLabel(str, Enum):
    """
    Enumeration representing tree relationships.

    Attributes:
        PARENT (str): Represents the parent edge.
        CHILD (str): Represents the child edge.
    """

    PARENT = "parent"
    CHILD = "child"


class TreeNode(Node):
    """
    A specialized node representing a node within a tree structure.

    Extends the basic node functionality with parent-child relationships.

    Attributes:
        parent (Node | None): The parent node. Defaults to None if the node has
            no parent.
    """

    parent: Node | None = Field(
        default=None,
        description="The parent node, as an instance of Node.",
    )

    @property
    def children(self) -> list[str]:
        """
        Retrieves the IDs of all child nodes.

        Returns:
            list[str]: A list containing the IDs of the child nodes.
        """
        if not self.parent:
            return list(self.related_nodes)
        else:
            return [
                node_id for node_id in self.related_nodes if node_id != self.parent.id_
            ]

    def relate_child(
        self,
        child: Node | list[Node],
        condition: Condition | None = None,
        bundle: bool = False,
    ) -> None:
        """
        Establishes a parent-child relationship between this node and the given
        child node(s).

        Args:
            child (Node | list[Node]): The child node or list of child nodes to
                be related.
            condition (Condition | None): The condition associated with the
                relationship, if any.
            bundle (bool): Indicates whether to bundle the relation into a
                single transaction. Defaults to False.
        """
        children = [child] if isinstance(child, Node) else child
        for _child in children:
            self.relate(
                _child,
                node_as="head",
                label=TreeLabel.PARENT,
                condition=condition,
                bundle=bundle,
            )
            if isinstance(_child, TreeNode):
                _child.parent = self

    def relate_parent(
        self,
        parent: Node,
        condition: Condition | None = None,
        bundle: bool = False,
    ) -> None:
        """
        Establishes a parent-child relationship between the given parent node
        and this node.

        Args:
            parent (Node): The parent node to be related.
            condition (Condition | None): The condition associated with the
                relationship, if any.
            bundle (bool): Indicates whether to bundle the relation into a
                single transaction. Defaults to False.
        """
        if self.parent:
            self.unrelate(self.parent)
        self.relate(
            parent,
            node_as="tail",
            label=TreeLabel.PARENT,
            condition=condition,
            bundle=bundle,
        )
        self.parent = parent

    def unrelate_parent(self):
        """
        Removes the parent relationship of this node.
        """
        self.unrelate(self.parent)
        self.parent = None

    def unrelate_child(self, child: Node | list[Node]):
        """
        Removes the child relationship between this node and the given child
        node(s).

        Args:
            child (Node | list[Node]): The child node or list of child nodes to
                be unrelated.
        """
        children: list[Node] = [child] if isinstance(child, Node) else child
        for _child in children:
            self.unrelate(_child)
            if isinstance(_child, TreeNode):
                _child.parent = None


class Tree(Graph):
    """
    Represents a tree structure, extending the graph with tree-specific
    functionalities.

    Manages parent-child relationships within the tree.

    Attributes:
        root (TreeNode | None): The root node of the tree. Defaults to None.
    """

    root: TreeNode | None = Field(
        default=None, description="The root node of the tree graph."
    )

    def relate_parent_child(
        self,
        parent: TreeNode,
        children: list[TreeNode],
        condition: Condition | None = None,
        bundle: bool = False,
    ) -> None:
        """
        Establishes parent-child relationships between the given parent node and
        child node(s).

        Args:
            parent (TreeNode): The parent node.
            children (list[TreeNode]): A list of child nodes.
            condition (Condition | None): The condition associated with the
                relationships, if any.
            bundle (bool): Indicates whether to bundle the relations into a
                single transaction. Defaults to False.
        """
        children = [children] if isinstance(children, TreeNode) else children
        for child in children:
            if child.id_ not in self.internal_nodes:
                self.add_node(child)

        for child in children:
            parent.relate_child(child, condition=condition, bundle=bundle)

        if parent.id_ not in self.internal_nodes:
            self.add_node(parent)

        if self.root is None:
            self.root = parent



from pydantic import Field

from lionagi.core.generic import Node
from lionagi.core.tool.tool import Tool


class ActionSelection(Node):
    action_kwargs: dict = Field(
        default_factory=dict,
        description="The arguments for the action",
        alias="action_arguments",
    )
    action: str = Field(
        "chat", description="The action to be performed", alias="action_type"
    )


class ActionNode(ActionSelection):
    tools: list[Tool] | Tool | None = Field(
        default_factory=list,
        description="The tools to be used in the action",
        alias="tool",
    )
    instruction: Node = Field(
        ..., description="The instruction for the action", alias="instruct"
    )



"""
This module provides the BaseStructure class, which represents a node with
internal edges and nodes, including methods for managing them.

The BaseStructure class inherits from the Node class and provides functionality
for adding, removing, and querying nodes and edges within the structure, as
well as checking if the structure is empty.
"""

from functools import singledispatchmethod
from typing import Any
from pydantic import Field

from lionagi.libs import convert, func_call
from .condition import Condition
from .edge import Edge
from .node import Node
from .action import ActionSelection
from lionagi.core.tool import Tool


class BaseStructure(Node):
    """
    Represents a node with internal edges and nodes, including methods for
    managing them.

    Provides functionality for adding, removing, and querying nodes and edges
    within the structure, as well as checking if the structure is empty.
    """

    internal_nodes: dict[str, Node] = Field(
        default_factory=dict,
        description="A dictionary of all nodes within the structure, keyed by \
            node ID.",
    )

    @property
    def internal_edges(self) -> dict[str, Edge]:
        """
        Gets all internal edges indexed by their ID.

        Returns:
            dict[str, Edge]: A dictionary of all internal edges within the
            structure.
        """
        edges = {}
        for i in self.internal_nodes.values():
            for _, edge in i.edges.items():
                if edge.id_ not in edges:
                    edges[edge.id_] = edge
        return edges

    @property
    def is_empty(self) -> bool:
        """
        Checks if the structure is empty (contains no nodes).

        Returns:
            bool: True if the structure has no nodes, False otherwise.
        """
        return len(self.internal_nodes) == 0

    def clear(self):
        """Clears all nodes and edges from the structure."""
        self.internal_nodes.clear()

    def get_node_edges(
        self,
        node: Node | str,
        node_as: str = "both",
        label: list | str = None,
    ) -> dict[str, list[Edge]]:
        """
        Retrieves edges associated with a specific node.

        Args:
            node (Node | str): The node or its ID to query for edges.
            node_as (str, optional): The role of the node in the edges.
                Defaults to "both".
                - "both" or "all": Retrieve edges where the node is either the
                  head or tail.
                - "head", "predecessor", "outgoing", "out", or "predecessors":
                  Retrieve edges where the node is the head.
                - "tail", "successor", "incoming", "in", or "successors":
                  Retrieve edges where the node is the tail.
            label (list | str, optional): Filter edges by label(s). Defaults to
                None.

        Returns:
            dict[str, list[Edge]]: A dictionary mapping related node IDs to
            lists of associated edges based on the specified criteria.

        Raises:
            ValueError: If an invalid role is specified for the node.
        """

        node = self.get_node(node)

        if node_as in ["all", "both"]:
            all_node_edges = {i: [] for i in node.related_nodes}

            for k, v in node.node_relations["points_to"].items():
                all_node_edges[k].append(v)

            for k, v in node.node_relations["pointed_by"].items():
                all_node_edges[k].append(v)

            for k, v in all_node_edges.items():
                all_node_edges[k] = convert.to_list(v, dropna=True, flatten=True)

            if label:
                for k, v in all_node_edges.items():
                    all_node_edges[k] = (
                        v
                        if v.label in convert.to_list(label, dropna=True, flatten=True)
                        else []
                    )

            return {k: v for k, v in all_node_edges.items() if len(v) > 0}

        elif node_as in ["head", "predecessor", "outgoing", "out", "predecessors"]:
            return node.node_relations["points_to"]

        elif node_as in ["tail", "successor", "incoming", "in", "successors"]:
            return node.node_relations["pointed_by"]

        else:
            raise ValueError(
                f"Invalid value for self_as: {node_as}, must be 'head' or 'tail'"
            )

    def relate_nodes(
        self,
        head: Node,
        tail: Node,
        condition: Condition | None = None,
        bundle=False,
        label=None,
        **kwargs,
    ):
        """
        Relates two nodes within the structure with an edge.

        Args:
            head (Node): The head node of the edge.
            tail (Node): The tail node of the edge.
            condition (Condition | None, optional): The condition associated
                with the edge. Defaults to None.
            label (optional): The label for the edge.
            **kwargs: Additional keyword arguments for the edge creation.

        Raises:
            ValueError: If there is an error adding the edge.
        """
        try:
            if isinstance(head, Tool) or isinstance(head, ActionSelection):
                raise ValueError(
                    f"type {type(tail)} should not be the head of the relationship, "
                    f"please switch position and attach it to the tail of the relationship"
                )

            if isinstance(tail, Tool) or isinstance(tail, ActionSelection):
                bundle = True

            if head.id_ not in self.internal_nodes:
                self.add_node(head)

            if tail.id_ not in self.internal_nodes:
                self.add_node(tail)

            head.relate(
                tail,
                node_as="head",
                condition=condition,
                label=label,
                bundle=bundle,
                **kwargs,
            )
        except Exception as e:
            raise ValueError(f"Error adding edge: {e}") from e

    def remove_edge(
        self, edge: Edge | str | list[str | Edge] | dict[str, Edge]
    ) -> bool:
        """
        Removes one or more edges from the structure.

        Args:
            edge: The edge(s) to be removed, specified as single edge, its ID,
                a list, or a dictionary of edges.

        Returns:
            bool: True if all specified edges were successfully removed, False
            otherwise.
        """
        if isinstance(edge, list):
            for i in edge:
                self._remove_edge(i)

        elif isinstance(edge, dict):
            for i in edge.values():
                self._remove_edge(i)

        elif isinstance(edge, (Edge, str)):
            self._remove_edge(edge)

    @singledispatchmethod
    def add_node(self, node: Any) -> None:
        """
        Method placeholder for adding a node. Will be implemented by register
        decorators.

        Args:
            node (Any): The node to be added.

        Raises:
            NotImplementedError: If the method is not implemented for the type
            of node.
        """
        raise NotImplementedError

    @add_node.register(Node)
    def _(self, node) -> None:
        if node.id_ not in self.internal_nodes:
            self.internal_nodes[node.id_] = node
        else:
            raise ValueError(f"Node {node.id_} already exists in structure.")

    @add_node.register(list)
    def _(self, node) -> None:
        for _node in node:
            self.add_node(_node)

    @add_node.register(dict)
    def _(self, node) -> None:
        self.add_node(list(node.values()))

    def get_node(self, node, default="undefined"):
        """
        Retrieves one or more nodes by ID, node instance, or a collection of
        IDs/nodes.

        Args:
            node: A single node, node ID, or a collection of nodes/IDs to
                retrieve.
            default: The default value to return if a node is not found.

        Returns:
            The node(s) corresponding to the input, or the default value for
            any not found.
        """
        if not isinstance(node, (list, dict, set)):
            return self._get_node(node, default=default)
        else:
            node = list(node) if isinstance(node, set) else node
            node = list(node.values()) if isinstance(node, dict) else node

            return func_call.lcall(node, self._get_node, default=default)

    @singledispatchmethod
    def remove_node(self, node: Any) -> bool:
        """
        Method placeholder for removing a node. Will be implemented by register
        decorators.

        Args:
            node (Any): The node to be removed.

        Returns:
            bool: True if the node was successfully removed, False otherwise.

        Raises:
            NotImplementedError: If the method is not implemented for the type
            of node.
        """
        return NotImplementedError

    @remove_node.register(Node)
    def _(self, node: Node) -> bool:
        return self._remove_node(node)

    @remove_node.register(str)
    def _(self, node: str) -> bool:
        return self._remove_node(node)

    @remove_node.register(list)
    def _(self, node: list[str | Node]) -> bool:
        for i in node:
            self.remove_node(i)

    @remove_node.register(dict)
    def _(self, node: dict[str, Node]) -> bool:
        self.remove_node(list(node.values()))

    def _pop_node(self, node: Node | str, default="undefined"):
        node = self.get_node(node, default=default)
        self.remove_node(node)
        return node

    @singledispatchmethod
    def pop_node(self, node: Any, default="undefined") -> Node:
        return NotImplementedError

    @pop_node.register(Node)
    def _(self, node: Node, default="undefined") -> Node:
        return self._pop_node(node, default=default)

    @pop_node.register(str)
    def _(self, node: str, default="undefined") -> Node:
        return self._pop_node(node, default=default)

    @pop_node.register(list)
    def _(self, node: list[str | Node], default="undefined") -> list[Node]:
        nodes = []
        for i in node:
            nodes.append(self.pop_node(i, default=default))
        return nodes

    @pop_node.register(dict)
    def _(self, node: dict[str, Node], default="undefined") -> list[Node]:
        return self.pop_node(list(node.values()), default=default)

    def _remove_edge(self, edge: Edge | str) -> bool:
        """Helper method to remove an edge by ID or edge instance.

        Args:
            edge (Edge | str): The edge or its ID to remove.

        Returns:
            bool: True if the edge was successfully removed, False otherwise.

        Raises:
            ValueError: If the edge is not found within the structure.
        """
        edge_id = edge.id_ if isinstance(edge, Edge) else edge
        if not edge_id in self.internal_edges:
            raise ValueError(f"Edge {edge_id} not found in structure.")

        edge = self.internal_edges[edge_id]

        head: Node = self.get_node(edge.head)
        tail: Node = self.get_node(edge.tail)

        head.unrelate(tail, edge=edge)

    def _get_node(self, node: Node | str, default="undefined") -> Node:
        node_id_ = node.id_ if isinstance(node, Node) else node
        if not node_id_ in self.internal_nodes:
            if default == "undefined":
                raise KeyError(f"Node {node_id_} not found in structure.")
            return default
        return self.internal_nodes[node_id_]

    def _remove_node(self, node: Node | str) -> bool:
        try:
            node = self.get_node(node)
            related_nodes = self.get_node(node.related_nodes)
            func_call.lcall(related_nodes, node.unrelate)
            self.internal_nodes.pop(node.id_)
            return True
        except Exception as e:
            raise ValueError(f"Error removing node: {e}") from e



import atexit
import contextlib
import logging
from collections import deque
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List

from lionagi.libs import SysUtil, convert, nested


# TODO: there should be a global data logger, under setting


@dataclass
class DLog:
    """
    Defines a log entry structure for data processing operations.

    This class encapsulates both the input to and output from a data processing operation,
    along with an automatically generated timestamp indicating when the log entry was
    created. It aims to standardize logging across applications for easier analysis
    and debugging.

    Attributes:
        input_data: The data input to the operation. Can be of any type.
        output_data: The data output by the operation. Can be of any type.
    """

    input_data: Any
    output_data: Any

    def serialize(self, *, flatten_: bool = True, sep: str = "[^_^]") -> dict[str, Any]:
        """Serializes the DLog instance into a dictionary with an added timestamp.

        Args:
            flatten_ (bool): If True, flattens dictionary inputs for serialization.
            sep (str): Separator used in flattening nested dictionaries.

        Returns:
            A dictionary representation of the DLog instance, including 'input_data',
            'output_data', and 'timestamp'.
        """
        log_dict = {}

        def _process_data(data, field):
            try:
                data = convert.to_str(data)
                if "{" not in data:
                    log_dict[field] = convert.to_str(data)

                else:
                    with contextlib.suppress(Exception):
                        data = convert.to_dict(data)

                    if isinstance(self.input_data, dict) and flatten_:
                        log_dict[field] = convert.to_str(nested.flatten(data, sep=sep))

                    else:
                        log_dict[field] = convert.to_str(data)

            except Exception as e:
                log_dict[field] = data
                logging.error(f"Error in processing {field} to str: {e}")

        _process_data(self.input_data, "input_data")
        _process_data(self.output_data, "output_data")

        log_dict["timestamp"] = SysUtil.get_timestamp()

        return log_dict

    @classmethod
    def deserialize(
        cls,
        *,
        input_str: str,
        output_str: str,
        unflatten_: bool = True,
        sep: str = "[^_^]",
    ) -> "DLog":
        """Deserializes log entries from string representations of input and output data.

        This method reconstructs a DLog instance from serialized string data, optionally
        unflattening nested dictionary structures if they were flattened during the
        serialization process. The method is particularly useful for reading logs from
        storage formats like JSON or CSV where data is represented as strings.

        Note:
            The separator '[^_^]' is reserved and should not be used within dictionary
            keys to ensure proper structure preservation during unflattening.

        Args:
            input_str: String representation of the input data.
            output_str: String representation of the output data.
            unflatten_ (bool): Indicates whether to unflatten the string data back into
                               nested dictionaries.
            sep (str): Separator used if unflattening is performed.

        Returns:
            An instance of DLog reconstructed from the provided string data.

        Raises:
            ValueError: If deserialization or unflattening fails due to incorrect data
                        format or separator issues.
        """

        def _process_data(data):
            if unflatten_:
                try:
                    return nested.unflatten(convert.to_dict(data), sep=sep)
                except:
                    return data
            else:
                return data

        input_data = _process_data(input_str)
        output_data = _process_data(output_str)

        return cls(input_data=input_data, output_data=output_data)


class DataLogger:
    """
    Manages logging for data processing activities within an application.

    This class handles the accumulation,
    structuring, and persistence of log entries.
    It supports exporting logs to disk in both CSV and JSON formats,
    with features for
    automatic log saving at program exit and customizable file naming.

    Attributes:
        persist_path: Path to the directory for saving log files.
        log: Container for log entries.
        filename: Base name for log files.
    """

    def __init__(
        self,
        persist_path: str | Path | None = None,
        log: List[Dict] | None = None,
        filename: str | None = None,
    ) -> None:
        """
        Initializes the DataLogger with optional custom settings for log storage.

        Args:
            persist_path: The file system path for storing log files. Defaults to 'data/logs/'.
            log: Initial log entries.
            filename: Base name for exported log files.
        """
        self.persist_path = Path(persist_path) if persist_path else Path("data/logs/")
        self.log = deque(log) if log else deque()
        self.filename = filename or "log"
        atexit.register(self.save_at_exit)

    def extend(self, logs) -> None:
        """
        Extends the log deque with multiple log entries.

        This method allows for bulk addition of log entries, which can be useful for
        importing logs from external sources or consolidating logs from different parts
        of an application.

        Args:
            logs: A list of log entries, each as a dictionary conforming to the log
                  structure (e.g., containing 'input_data', 'output_data', etc.).
        """
        if len(logs) > 0:
            log1 = convert.to_list(self.log)
            log1.extend(convert.to_list(logs))
            self.log = deque(log1)

    def append(self, *, input_data: Any, output_data: Any) -> None:
        """
        Appends a new log entry from provided input and output data.

        Args:
            input_data: Input data to the operation.
            output_data: Output data from the operation.
        """
        log_entry = DLog(input_data=input_data, output_data=output_data)
        self.log.append(log_entry)

    def to_csv_file(
        self,
        filename: str = "log.csv",
        *,
        dir_exist_ok: bool = True,
        timestamp: bool = True,
        time_prefix: bool = False,
        verbose: bool = True,
        clear: bool = True,
        flatten_=True,
        sep="[^_^]",
        index=False,
        random_hash_digits=3,
        **kwargs,
    ) -> None:
        """Exports log entries to a CSV file with customizable options.

        Args:
            filename: Filename for the exported CSV. Defaults to 'log.csv'.
            dir_exist_ok: If True, allows writing to an existing directory.
            timestamp: If True, appends a timestamp to the filename.
            time_prefix: If True, places the timestamp prefix before the filename.
            verbose: If True, prints a message upon successful save.
            clear: If True, clears the log deque after saving.
            flatten_: If True, flattens dictionary data for serialization.
            sep: Separator for flattening nested dictionaries.
            index: If True, includes an index column in the CSV.
            **kwargs: Additional arguments for DataFrame.to_csv().
        """

        if not filename.endswith(".csv"):
            filename += ".csv"

        filepath = SysUtil.create_path(
            self.persist_path,
            filename,
            timestamp=timestamp,
            dir_exist_ok=dir_exist_ok,
            time_prefix=time_prefix,
            random_hash_digits=random_hash_digits,
        )
        try:
            logs = [log.serialize(flatten_=flatten_, sep=sep) for log in self.log]
            df = convert.to_df(convert.to_list(logs, flatten=True))
            df.to_csv(filepath, index=index, **kwargs)
            if verbose:
                print(f"{len(self.log)} logs saved to {filepath}")
            if clear:
                self.log.clear()
        except Exception as e:
            raise ValueError(f"Error in saving to csv: {e}") from e

    def to_json_file(
        self,
        filename: str = "log.json",
        *,
        dir_exist_ok: bool = True,
        timestamp: bool = True,
        time_prefix: bool = False,
        verbose: bool = True,
        clear: bool = True,
        flatten_=True,
        sep="[^_^]",
        index=False,
        random_hash_digits=3,
        **kwargs,
    ) -> None:
        """Exports log entries to a JSON file with customizable options.

        Args:
            filename: Filename for the exported JSON. Defaults to 'log.json'.
            dir_exist_ok: If True, allows writing to an existing directory.
            timestamp: If True, appends a timestamp to the filename.
            time_prefix: If True, places the timestamp prefix before the filename.
            verbose: If True, prints a message upon successful save.
            clear: If True, clears the log deque after saving.
            flatten_: If True, flattens dictionary data for serialization.
            sep: Separator for flattening nested dictionaries.
            index: If True, includes an index in the JSON.
            **kwargs: Additional arguments for DataFrame.to_json().
        """
        if not filename.endswith(".json"):
            filename += ".json"

        filepath = SysUtil.create_path(
            self.persist_path,
            filename,
            timestamp=timestamp,
            dir_exist_ok=dir_exist_ok,
            time_prefix=time_prefix,
            random_hash_digits=random_hash_digits,
        )

        try:
            logs = [log.serialize(flatten_=flatten_, sep=sep) for log in self.log]
            df = convert.to_df(convert.to_list(logs, flatten=True))
            df.to_json(filepath, index=index, **kwargs)
            if verbose:
                print(f"{len(self.log)} logs saved to {filepath}")
            if clear:
                self.log.clear()
        except Exception as e:
            raise ValueError(f"Error in saving to csv: {e}") from e

    def save_at_exit(self):
        """
        Registers an at-exit handler to ensure that any unsaved logs are automatically
        persisted to a file upon program termination. This safeguard helps prevent the
        loss of log data due to unexpected shutdowns or program exits.

        The method is configured to save the logs to a CSV file, named
        'unsaved_logs.csv', which is stored in the designated persisting directory. This
        automatic save operation is triggered only if there are unsaved logs present at
        the time of program exit.

        Note: This method does not clear the logs after saving, allowing for the
        possibility of manual.py review or recovery after the program has terminated.
        """
        if self.log:
            self.to_csv_file("unsaved_logs.csv", clear=False)



from typing import Any

from pydantic import field_serializer
from lionagi.core.generic import Node


class Tool(Node):
    """
    Represents a tool, extending BaseNode with specific functionalities and configurations.

    Attributes:
        func: The main function or capability of the tool.
        schema_: An optional schema defining the structure and constraints of data the tool works with.
        manual: Optional documentation or manual for using the tool.
        parser: An optional parser associated with the tool for data processing or interpretation.
    """

    func: Any
    schema_: dict | None = None
    manual: Any | None = None
    parser: Any | None = None

    @field_serializer("func")
    def serialize_func(self, func):
        return func.__name__


TOOL_TYPE = bool | Tool | str | list[Tool | str | dict] | dict
